# coding: utf8


import abc
import os.path as path

import numpy as np
from pandas.io import parsers

from clinica.pipelines.machine_learning import base
import clinica.pipelines.machine_learning.voxel_based_io as vbio
import clinica.pipelines.machine_learning.vertex_based_io as vtxbio
import clinica.pipelines.machine_learning.region_based_io as rbio
import clinica.pipelines.machine_learning.tsv_based_io as tbio
import clinica.pipelines.machine_learning.ml_utils as utils


__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Jorge Samper-Gonzalez", "Simona Bottani"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


class CAPSInput(base.MLInput):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, precomputed_kernel=None):
        """

        Args:
            caps_directory:
            subjects_visits_tsv:
            diagnoses_tsv:
            group_id:
            image_type: 'T1', 'fdg', 'av45', 'pib' or 'flute'
            precomputed_kernel:
        """

        self._caps_directory = caps_directory
        self._group_id = group_id
        self._image_type = image_type
        self._images = None
        self._x = None
        self._y = None
        self._kernel = None

        subjects_visits = parsers.read_csv(subjects_visits_tsv, sep='\t')
        if list(subjects_visits.columns.values) != ['participant_id', 'session_id']:
            raise Exception('Subjects and visits file is not in the correct format.')
        self._subjects = list(subjects_visits.participant_id)
        self._sessions = list(subjects_visits.session_id)

        diagnoses = parsers.read_csv(diagnoses_tsv, sep='\t')
        if 'diagnosis' not in list(diagnoses.columns.values):
            raise Exception('Diagnoses file is not in the correct format.')
        self._diagnoses = list(diagnoses.diagnosis)

        if image_type not in ['T1', 'fdg', 'av45', 'pib', 'flute', 'dwi']:
            raise Exception("Incorrect image type. It must be one of the values 'T1', 'fdg', 'av45', 'pib', 'flute' or 'dwi'")

        if precomputed_kernel is not None:
            if type(precomputed_kernel) == np.ndarray:
                if precomputed_kernel.shape == (len(self._subjects), len(self._subjects)):
                    self._kernel = precomputed_kernel
                else:
                    raise Exception("""Precomputed kernel provided is not in the correct format.
                    It must be a numpy.ndarray object with number of rows and columns equal to the number of subjects,
                    or a filename to a numpy txt file containing an object with the described format.""")
            elif type(precomputed_kernel == str):
                self._kernel = np.loadtxt(precomputed_kernel)
            else:
                raise Exception("""Precomputed kernel provided is not in the correct format.
                It must be a numpy.ndarray object with number of rows and columns equal to the number of subjects,
                or a filename to a numpy txt file containing an object with the described format.""")

    @abc.abstractmethod
    def get_images(self):
        """

        Returns: a list of filenames

        """
        pass

    @abc.abstractmethod
    def get_x(self):
        """

        Returns: a numpy 2d-array.

        """
        pass

    def get_y(self):
        """

        Returns: a list of integers. Each integer represents a class.

        """
        if self._y is not None:
            return self._y

        unique = sorted(list(set(self._diagnoses)))
        self._y = np.array([unique.index(x) for x in self._diagnoses])
        return self._y

    def get_kernel(self, kernel_function=utils.gram_matrix_linear, recompute_if_exists=False):
        """

        Returns: a numpy 2d-array.

        """
        if self._kernel is not None and not recompute_if_exists:
            return self._kernel

        if self._x is None:
            self.get_x()

        print("Computing kernel ...")
        self._kernel = kernel_function(self._x)
        print("Kernel computed")
        return self._kernel

    def save_kernel(self, output_dir):
        """

        Args:
            output_dir:

        Returns:

        """
        if self._kernel is not None:
            filename = path.join(output_dir, 'kernel.txt')
            np.savetxt(filename, self._kernel)
            return filename
        raise Exception("Unable to save the kernel. Kernel must have been computed before.")

    @abc.abstractmethod
    def save_weights_as_nifti(self, weights, output_dir):
        pass


class CAPSVoxelBasedInput(CAPSInput):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, fwhm=0,
                 modulated="on", pvc=None, mask_zeros=True, precomputed_kernel=None):
        """

        Args:
            caps_directory:
            subjects_visits_tsv:
            diagnoses_tsv:
            group_id:
            image_type: 'T1', 'fdg', 'av45', 'pib' or 'flute'
            fwhm:
            modulated:
            mask_zeros:
            precomputed_kernel:
        """

        super(CAPSVoxelBasedInput, self).__init__(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                  image_type, precomputed_kernel=precomputed_kernel)

        self._fwhm = fwhm
        self._modulated = modulated
        self._pvc = pvc
        self._mask_zeros = mask_zeros
        self._orig_shape = None
        self._data_mask = None

        if modulated not in ['on', 'off']:
            raise Exception("Incorrect modulation parameter. It must be one of the values 'on' or 'off'")

    def get_images(self):
        """

        Returns: a list of filenames

        """
        if self._images is not None:
            return self._images

        if self._image_type == 'T1':
            fwhm = '' if self._fwhm == 0 else '_fwhm-%dmm' % int(self._fwhm)

            self._images = [path.join(self._caps_directory, 'subjects', self._subjects[i], self._sessions[i],
                                      't1/spm/dartel/group-' + self._group_id,
                                      '%s_%s_T1w_segm-graymatter_space-Ixi549Space_modulated-%s%s_probability.nii.gz'
                                      % (self._subjects[i], self._sessions[i], self._modulated, fwhm))
                            for i in range(len(self._subjects))]
        else:
            pvc = '' if self._pvc is None else '_pvc-%s' % self._pvc
            fwhm = '' if self._fwhm == 0 else '_fwhm-%dmm' % int(self._fwhm)
            suvr = 'pons' if self._image_type == 'fdg' else 'cerebellumPons'

            self._images = [path.join(self._caps_directory, 'subjects', self._subjects[i], self._sessions[i],
                                      'pet/preprocessing/group-' + self._group_id,
                                      '%s_%s_task-rest_acq-%s_pet_space-Ixi549Space%s_suvr-%s_mask-brain%s_pet.nii.gz'
                                      % (self._subjects[i], self._sessions[i], self._image_type, pvc, suvr, fwhm))
                            for i in range(len(self._subjects))]

        for image in self._images:
            if not path.exists(image):
                raise Exception("File %s doesn't exists." % image)

        return self._images

    def get_x(self):
        """

        Returns: a numpy 2d-array.

        """
        if self._x is not None:
            return self._x

        print('Loading ' + str(len(self.get_images())) + ' subjects')
        self._x, self._orig_shape, self._data_mask = vbio.load_data(self._images, mask=self._mask_zeros)
        print('Subjects loaded')

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):

        if self._images is None:
            self.get_images()

        output_filename = path.join(output_dir, 'weights.nii.gz')
        data = vbio.revert_mask(weights, self._data_mask, self._orig_shape)
        vbio.weights_to_nifti(data, self._images[0], output_filename)


class CAPSRegionBasedInput(CAPSInput):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, atlas,
                 pvc=None, precomputed_kernel=None):
        """

        Args:
            caps_directory:
            subjects_visits_tsv:
            diagnoses_tsv:
            group_id:
            image_type: 'T1', 'fdg', 'av45', 'pib' or 'flute'
            atlas:
            precomputed_kernel:
        """

        super(CAPSRegionBasedInput, self).__init__(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                   image_type, precomputed_kernel=precomputed_kernel)

        self._atlas = atlas
        self._pvc = pvc
        self._orig_shape = None
        self._data_mask = None

        if atlas not in ['AAL2', 'Neuromorphometrics', 'AICHA', 'LPBA40', 'Hammers']:
            raise Exception("Incorrect atlas name. It must be one of the values 'AAL2', 'Neuromorphometrics', 'AICHA', 'LPBA40', 'Hammers' ")

    def get_images(self):
        """

        Returns: a list of filenames

        """
        if self._images is not None:
            return self._images

        if self._image_type == 'T1':
            self._images = [path.join(self._caps_directory, 'subjects', self._subjects[i], self._sessions[i],
                                      't1/spm/dartel/group-' + self._group_id,
                                      'atlas_statistics/', '%s_%s_T1w_space-%s_map-graymatter_statistics.tsv'
                                      % (self._subjects[i], self._sessions[i], self._atlas))
                            for i in range(len(self._subjects))]
        else:
            pvc = '' if self._pvc is None else '_pvc-%s' % self._pvc
            suvr = 'pons' if self._image_type == 'fdg' else 'cerebellumPons'

            self._images = [path.join(self._caps_directory, 'subjects', self._subjects[i], self._sessions[i],
                                      'pet/preprocessing/group-' + self._group_id, 'atlas_statistics',
                                      '%s_%s_task-rest_acq-%s_pet_space-%s%s_suvr-%s_statistics.tsv'
                                      % (self._subjects[i], self._sessions[i], self._image_type, self._atlas, pvc, suvr))
                            for i in range(len(self._subjects))]

        for image in self._images:
            if not path.exists(image):
                raise Exception("File %s doesn't exists." % image)

        return self._images

    def get_x(self):
        """

        Returns: a numpy 2d-array.

        """
        if self._x is not None:
            return self._x

        print('Loading ' + str(len(self.get_images())) + ' subjects')
        self._x = rbio.load_data(self._images, self._subjects)
        print('Subjects loaded')

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):
        """

        Args:
            weights:
            output_dir:

        Returns:

        """

        output_filename = path.join(output_dir, 'weights.nii.gz')
        rbio.weights_to_nifti(weights, self._atlas, output_filename)


class CAPSVertexBasedInput(CAPSInput):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, fwhm, image_type, precomputed_kernel=None):
        super(CAPSVertexBasedInput, self).__init__(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                   image_type, precomputed_kernel)
        self._fwhm = fwhm
        self._image_type = image_type
        self._caps_directory = caps_directory

    def get_images(self):
        import os
        """
        returns list of filnames
        """

        if self._images is not None:
            return self._images

        if self._image_type == 'fdg' and self._images is None:
            self._images = []
            hemi = ['lh', 'rh']
            for i in range(len(self._subjects)):
                self._images.append([os.path.join(self._caps_directory, 'subjects', self._subjects[i], self._sessions[i], 'pet',
                                                  'surface', self._subjects[i] + '_' + self._sessions[i]
                                                  + '_task-rest_acq-fdg_pet_space-fsaverage_suvr-pons_pvc-iy_hemi-' + h
                                                  + '_fwhm-' + str(self._fwhm) + '_projection.mgh') for h in hemi])
            missing_files = []
            missing_files_string_error = ''
            for img in self._images:
                for side in img:
                    if not os.path.exists(side):
                        missing_files.append(side)
                        missing_files_string_error += side + '\n'
            if len(missing_files) > 0:
                raise IOError('Could not find the following files : \n' + missing_files_string_error
                              + '\n' + str(len(missing_files)) + ' files missing')
        return self._images

    def get_x(self):
        from clinica.utils.stream import cprint
        """
        Returns numpy 2D array
        """

        if self._x is not None:
            return self._x

        cprint('Loading ' + str(len(self.get_images())) + ' subjects')
        self._x = vtxbio.load_data(self._images)
        cprint(str(len(self._x)) + ' subjects loaded')
        return self._x

    def save_weights_as_datasurface(self, weights, output_dir):
        import numpy as np
        import nibabel as nib
        import os

        if self._images is None:
            self.get_images()

        sample = nib.load(self._images[0][0])

        infinite_norm = np.max(np.abs(weights))

        left_hemi_data = np.atleast_3d(np.divide(weights[:np.int(weights.size / 2)], infinite_norm))
        left_hemi_mgh = nib.MGHImage(left_hemi_data, affine=sample.affine, header=sample.header)
        nib.save(left_hemi_mgh, os.path.join(output_dir, 'weights_lh.mgh'))

        right_hemi_data = np.atleast_3d(np.divide(weights[np.int(weights.size/2):], infinite_norm))
        right_hemi_mgh = nib.MGHImage(right_hemi_data, affine=sample.affine, header=sample.header)
        nib.save(right_hemi_mgh, os.path.join(output_dir, 'weights_rh.mgh'))
        pass

    def save_weights_as_nifti(self, weights, output_dir):
        pass


class CAPSTSVBasedInput(CAPSInput):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, atlas, dataset,
                 pvc=None, precomputed_kernel=None):
        """

        Args:
            caps_directory:
            subjects_visits_tsv:
            diagnoses_tsv:
            group_id:
            image_type: 'T1', 'fdg', 'av45', 'pib' or 'flute'
            atlas:
            precomputed_kernel:
        """

        super(CAPSTSVBasedInput, self).__init__(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                image_type, precomputed_kernel)

        self._atlas = atlas
        self._pvc = pvc
        self._dataset = dataset

        self._orig_shape = None
        self._data_mask = None

        if atlas not in ['AAL2', 'Neuromorphometrics', 'AICHA', 'LPBA40', 'Hammers']:
            raise Exception("Incorrect atlas name. It must be one of the values 'AAL2', 'Neuromorphometrics', 'AICHA', 'LPBA40', 'Hammers' ")

    def get_images(self):
        """

        Returns: string

        """

        pass

    def get_x(self):
        """

        Returns: a numpy 2d-array.

        """

        # if self._x is not None:
        #    return self._x

        print('Loading TSV subjects')
        string = str('group-' + self._group_id + '_T1w_space-' + self._atlas + '_map-graymatter')

        self._x = tbio.load_data(string, self._caps_directory, self._subjects, self._sessions, self._dataset)

        print('Subjects loaded')

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):
        """

        Args:
            weights:
            output_dir:

        Returns:

        """

        # output_filename = path.join(output_dir, 'weights.nii.gz')

        # rbio.weights_to_nifti(weights, self._atlas, output_filename)
        pass


class CAPSVoxelBasedInputREGSVM(CAPSInput):

    def __init__(self, caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id, image_type, fwhm=0,
                 modulated="on", pvc=None, mask_zeros=True, precomputed_kernel=None):
        """

        Args:
            caps_directory:
            subjects_visits_tsv:
            diagnoses_tsv:
            group_id:
            image_type: 'T1', 'fdg', 'av45', 'pib' or 'flute'
            fwhm:
            modulated:
            mask_zeros:
            precomputed_kernel:
        """

        super(CAPSVoxelBasedInputREGSVM, self).__init__(caps_directory, subjects_visits_tsv, diagnoses_tsv, group_id,
                                                        image_type, precomputed_kernel=precomputed_kernel)

        self._fwhm = fwhm
        self._modulated = modulated
        self._pvc = pvc
        self._mask_zeros = mask_zeros
        self._orig_shape = None
        self._data_mask = None

        if modulated not in ['on', 'off']:
            raise Exception("Incorrect modulation parameter. It must be one of the values 'on' or 'off'")

    def get_images(self):
        """

        Returns: a list of filenames

        """
        if self._images is not None:
            return self._images

        if self._image_type == 'T1':
            fwhm = '' if self._fwhm == 0 else '_fwhm-%dmm' % int(self._fwhm)

            self._images = [path.join(self._caps_directory,
                                      'regul_%s_%s_T1w_segm-graymatter_space-Ixi549Space_modulated-%s%s_probability.nii'
                                      % (self._subjects[i], self._sessions[i], self._modulated, fwhm))
                            for i in range(len(self._subjects))]
        else:
            pvc = '' if self._pvc is None else '_pvc-%s' % self._pvc
            fwhm = '' if self._fwhm == 0 else '_fwhm-%dmm' % int(self._fwhm)
            suvr = 'pons' if self._image_type == 'fdg' else 'cerebellumPons'
            self._images = [path.join(self._caps_directory, 'subjects', self._subjects[i], self._sessions[i],
                                      'pet/preprocessing/group-' + self._group_id,
                                      '%s_%s_task-rest_acq-%s_pet_space-Ixi549Space%s_suvr-%s_mask-brain%s_pet.nii.gz'
                                      % (self._subjects[i], self._sessions[i], self._image_type, pvc, suvr, fwhm))
                            for i in range(len(self._subjects))]

        for image in self._images:
            if not path.exists(image):
                raise Exception("File %s doesn't exists." % image)

        return self._images

    def get_x(self):
        """

        Returns: a numpy 2d-array.

        """
        if self._x is not None:
            return self._x

        print('Loading ' + str(len(self.get_images())) + ' subjects')
        self._x, self._orig_shape, self._data_mask = vbio.load_data(self._images, mask=self._mask_zeros)
        print('Subjects loaded')

        return self._x

    def save_weights_as_nifti(self, weights, output_dir):

        if self._images is None:
            self.get_images()

        output_filename = path.join(output_dir, 'weights.nii.gz')
        data = vbio.revert_mask(weights, self._data_mask, self._orig_shape)
        vbio.weights_to_nifti(data, self._images[0], output_filename)



# coding: utf8

INDEX_TISSUE_MAP = {
    1: 'graymatter',
    2: 'whitematter',
    3: 'csf',
    4: 'bone',
    5: 'softtissue',
    6: 'background'
}


def check_spm_home():
    """Check and get SPM_HOME environment variable if present."""
    import os
    import platform
    from colorama import Fore
    from .check_dependency import check_environment_variable
    from .exceptions import ClinicaMissingDependencyError

    spm_home = check_environment_variable('SPM_HOME', 'SPM')

    spm_standalone_home = os.environ.get('SPMSTANDALONE_HOME', '')
    if spm_standalone_home:
        if not os.path.isdir(spm_standalone_home):
            raise ClinicaMissingDependencyError(
                '%s\n[Error] The SPMSTANDALONE_HOME environment variable you gave is not a folder (content: %s).%s'
                % (Fore.RED, spm_standalone_home, Fore.RESET))
        if platform.system() == 'Darwin':
            spm_home = os.path.join(spm_standalone_home, 'spm12.app', 'Contents', 'MacOS', 'spm12_mcr')
        else:
            spm_home = os.path.join(spm_standalone_home, 'spm12_mcr')

    return spm_home


def get_tpm():
    """
    Get Tissue Probability Map (TPM) from SPM
    Args
        None
    Returns
        TPM.nii from SPM
    """
    import os
    from glob import glob
    from os.path import join

    spm_home = os.getenv('SPM_HOME')

    if not spm_home:
        # Try MCR to get a hint on SPM location
        spm_home = os.getenv('SPMSTANDALONE_HOME')

    if not spm_home:
        raise RuntimeError('Could not determine location of your SPM installation. '
                           + 'Neither $SPM_HOME or $SPMSTANDALONE_HOME are present '
                             'in your environment')

    tpm_file_glob = glob(join(spm_home, '**/TPM.nii'), recursive=True)
    if len(tpm_file_glob) == 0:
        raise RuntimeError('No file found for TPM.nii in your $SPM_HOME in ' + spm_home)
    if len(tpm_file_glob) > 1:
        error_str = 'Multiple files found for TPM.nii in your SPM_HOME ' + spm_home + ':'
        for file in tpm_file_glob:
            error_str += '\n\t' + file
        raise RuntimeError(error_str)
    return tpm_file_glob[0]



# coding: utf-8

"""
 Module for converting AV45 and Florbetaben PET of ADNI
"""

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Sabrina Fontanella"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_av45_fbb_pet(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert AV-45 and Florbetaben PET images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import t1_pet_paths_to_bids
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of AV45 and Florbetaben PET images. Output will be stored in %s.' % path.join(dest_dir, 'conversion_info'))
    images = compute_av45_fbb_pet_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of AV45 and Florbetaben PET images found. Exporting images into BIDS ...')
    t1_pet_paths_to_bids(images, dest_dir, 'av45_fbb')
    cprint(Fore.GREEN + 'AV45 and Florbetaben PET conversion done.' + Fore.RESET)


def compute_av45_fbb_pet_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute the paths to the AV45 and Florbetaben PET images and store them in a tsv file

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns:
        images: a dataframe with all the paths to the PET images that will be converted into BIDS

    """

    import pandas as pd
    import os
    import operator
    from os import path
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars, find_image_path
    from clinica.utils.stream import cprint
    from functools import reduce

    pet_amyloid_col = ['Phase', 'Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date', 'Study_ID', 'Series_ID',
                       'Image_ID', 'Original', 'Tracer']
    pet_amyloid_df = pd.DataFrame(columns=pet_amyloid_col)

    # Loading needed .csv files
    av45qc = pd.read_csv(path.join(csv_dir, 'AV45QC.csv'), sep=',', low_memory=False)
    amyqc = pd.read_csv(path.join(csv_dir, 'AMYQC.csv'), sep=',', low_memory=False)
    pet_meta_list = pd.read_csv(path.join(csv_dir, 'PET_META_LIST.csv'), sep=',', low_memory=False)

    for subj in subjs_list:

        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list['Subject'] == subj]

        if subject_pet_meta.shape[0] < 1:
            # TODO Log somewhere subjects without Amyloid PET images
            # cprint('No Amyloid PET images metadata for subject - ' + subj)
            continue

        # QC for AV45 PET images for ADNI 1, GO and 2
        av45_qc_subj = av45qc[(av45qc.PASS == 1) & (av45qc.RID == int(subj[-4:]))]

        # QC for Amyloid PET images for ADNI 3
        amy_qc_subj = amyqc[(amyqc.SCANQLTY == 1) & (amyqc.RID == int(subj[-4:]))]
        amy_qc_subj.insert(0, 'EXAMDATE', amy_qc_subj.SCANDATE.to_list())

        # Concatenating visits in both QC files
        amyloid_qc_subj = pd.concat([av45_qc_subj, amy_qc_subj], axis=0, ignore_index=True, sort=False)

        for visit in list(amyloid_qc_subj.VISCODE2.unique()):
            amyloid_qc_visit = amyloid_qc_subj[amyloid_qc_subj.VISCODE2 == visit]

            # TODO Check
            # If there are several scans for a timepoint we keep image acquired last (higher LONIUID)
            amyloid_qc_visit = amyloid_qc_visit.sort_values("LONIUID", ascending=False)

            qc_visit = amyloid_qc_visit.iloc[0]

            # Corresponding LONI image ID for original scan in PET Meta List
            int_image_id = int(qc_visit.LONIUID[1:])

            original_pet_meta = subject_pet_meta[(subject_pet_meta['Orig/Proc'] == 'Original')
                                                 & (subject_pet_meta['Image ID'] == int_image_id)
                                                 & (subject_pet_meta.Sequence.map(
                                                        lambda s: (s.lower().find('early') < 0)))]

            # If no corresponding Amyloid PET metadata for an original image,
            # take scan at the same date
            if original_pet_meta.shape[0] < 1:
                original_pet_meta = subject_pet_meta[(subject_pet_meta['Orig/Proc'] == 'Original')
                                                     & (subject_pet_meta['Scan Date'] == qc_visit.EXAMDATE)
                                                     & (subject_pet_meta.Sequence.map(
                                                            lambda s: (s.lower().find('early') < 0)))]

                if original_pet_meta.shape[0] < 1:
                    # TODO Log somewhere QC visits without image metadata
                    cprint('No Amyloid PET images metadata for subject - ' + subj + ' for visit ' + qc_visit.VISCODE2)
                    continue

            original_image = original_pet_meta.iloc[0]

            # To determine type of amyloid PET tracer we find the
            # Coreg, Avg, Std Img and Vox Siz, Uniform Resolution image
            # with the same Series ID of the original image
            final_pet_meta = subject_pet_meta[(subject_pet_meta.Sequence.map(
                lambda x: (x.find('Coreg, Avg, Std Img and Vox Siz, Uniform Resolution') > 0)))
                                              & (subject_pet_meta['Series ID'] == original_image['Series ID'])]

            if final_pet_meta.shape[0] < 1:
                final_pet_meta = subject_pet_meta[(subject_pet_meta.Sequence.map(
                    lambda x: (x.find('Coreg, Avg, Std Img and Vox Siz, Uniform Resolution') > 0)))
                                                     & (subject_pet_meta['Scan Date'] == original_image['Scan Date'])]
                if final_pet_meta.shape[0] < 1:
                    # TODO Log
                    cprint('No "Coreg, Avg, Std Img and Vox Siz, Uniform Resolution" Amyloid PET image metadata for subject'
                           ' ' + subj + ' for visit ' + qc_visit.VISCODE2)
                    continue

            processed_sequence = final_pet_meta.iloc[0].Sequence

            if processed_sequence.startswith('AV45'):
                tracer = 'AV45'
            elif processed_sequence.startswith('FBB'):
                tracer = 'FBB'
            else:
                # TODO Log
                cprint('Unknown tracer for Amyloid PET image metadata for subject ' + subj +
                       ' for visit ' + qc_visit.VISCODE2)
                continue

            # Co-registered and Averaged image with the same Series ID of the original image
            averaged_pet_meta = subject_pet_meta[(subject_pet_meta['Sequence'] == '%s Co-registered, Averaged' % tracer)
                                                 & (subject_pet_meta['Series ID'] == original_image['Series ID'])]

            # If an explicit Co-registered, Averaged image does not exist,
            # the original image is already in that preprocessing stage.

            if averaged_pet_meta.shape[0] < 1:
                sel_image = original_image
                original = True
            else:
                sel_image = averaged_pet_meta.iloc[0]
                original = False

            visit = sel_image.Visit
            sequence = replace_sequence_chars(sel_image.Sequence)
            date = sel_image['Scan Date']
            study_id = sel_image['Study ID']
            series_id = sel_image['Series ID']
            image_id = sel_image['Image ID']

            row_to_append = pd.DataFrame(
                [[qc_visit.Phase, subj, qc_visit.VISCODE2, str(visit), sequence, date, str(study_id), str(series_id),
                  str(image_id), original, tracer]],
                columns=pet_amyloid_col)
            pet_amyloid_df = pet_amyloid_df.append(row_to_append, ignore_index=True)

    # TODO check for new exceptions in ADNI3
    # Exceptions
    # ==========
    conversion_errors = [  # Eq_1
                         ('128_S_2220', 'm48')]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((pet_amyloid_df.Subject_ID == conv_error[0])
                             & (pet_amyloid_df.VISCODE == conv_error[1]))

    if error_indices:
        indices_to_remove = pet_amyloid_df.index[reduce(operator.or_, error_indices, False)]
        pet_amyloid_df.drop(indices_to_remove, inplace=True)

    # DONE - Make a function reusing this code for different modalities
    # TODO check if it works properly

    images = find_image_path(pet_amyloid_df, source_dir, 'Amyloid', 'I', 'Image_ID')

    amyloid_csv_path = path.join(dest_dir, 'conversion_info')
    if not os.path.exists(amyloid_csv_path):
        os.mkdir(amyloid_csv_path)
    images.to_csv(path.join(amyloid_csv_path, 'amyloid_pet_paths.tsv'), sep='\t', index=False)

    return images


def check_exceptions(bids_dir):
    from os import path
    import pandas as pd
    from glob import glob

    amyloid_paths = pd.read_csv(path.join(bids_dir, 'conversion_info', 'amyloid_pet_paths.tsv'), sep='\t')

    av45_paths = amyloid_paths[amyloid_paths.Tracer == 'AV45']

    av45_paths = av45_paths[av45_paths.Path.notnull()]

    av45_paths['BIDS_SubjID'] = ['sub-ADNI' + s.replace('_', '') for s in av45_paths.Subject_ID.to_list()]
    av45_paths['BIDS_Session'] = ['ses-' + s.replace('bl', 'm00').upper() for s in av45_paths.VISCODE.to_list()]

    count = 0

    for r in av45_paths.iterrows():
        image = r[1]
        image_dir = path.join(bids_dir, image.BIDS_SubjID, image.BIDS_Session, 'pet')
        image_pattern = path.join(image_dir, '%s_%s_*av45*' % (image.BIDS_SubjID, image.BIDS_Session))
        files_list = glob(image_pattern)
        if not files_list:
            print("No images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            count += 1

        elif len(files_list) > 1:
            print("Too many images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            print(files_list)

    print(count)

    fbb_paths = amyloid_paths[amyloid_paths.Tracer == 'FBB']

    fbb_paths = fbb_paths[fbb_paths.Path.notnull()]

    fbb_paths['BIDS_SubjID'] = ['sub-ADNI' + s.replace('_', '') for s in fbb_paths.Subject_ID.to_list()]
    fbb_paths['BIDS_Session'] = ['ses-' + s.replace('bl', 'm00').upper() for s in fbb_paths.VISCODE.to_list()]

    count = 0

    for r in fbb_paths.iterrows():
        image = r[1]
        image_dir = path.join(bids_dir, image.BIDS_SubjID, image.BIDS_Session, 'pet')
        image_pattern = path.join(image_dir, '%s_%s_*fbb*' % (image.BIDS_SubjID, image.BIDS_Session))
        files_list = glob(image_pattern)
        if not files_list:
            print("No images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            count += 1

        elif len(files_list) > 1:
            print("Too many images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            print(files_list)

    print(count)



# coding: utf8

import clinica.engine as ce

__author__ = "Sabrina Fontanella"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Simona Bottani"
__email__ = "simona.bottani@icm-institute.org"
__status__ = "Completed"


class OasisToBidsCLI(ce.CmdParser):
    def define_name(self):
        """Define the sub-command name to run this pipelines.
        """
        self._name = 'oasis-to-bids'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Convert OASIS (http://oasis-brains.org/) into BIDS.'

    def define_options(self):
        """Define the sub-command arguments
        """
        self._args.add_argument("dataset_directory",
                                help='Path to the OASIS images directory.')
        self._args.add_argument("clinical_data_directory",
                                help='Path to the OASIS clinical data directory.')
        self._args.add_argument("bids_directory",
                                help='Path to the BIDS directory.')

    def run_command(self, args):
        from clinica.iotools.converters.oasis_to_bids.oasis_to_bids import OasisToBids
        from clinica.iotools.converter_utils import check_bin
        from clinica.utils.stream import cprint
        from colorama import Fore
        import sys

        missing_bin = check_bin('mri_convert')
        if missing_bin:
            cprint(Fore.RED + 'mri_convert from FreeSurfer is required.'
                   + ' Install it and re-run the converter.' + Fore.RESET)
            cprint('Exiting clinica...')
            sys.exit()
        oasis_to_bids = OasisToBids()

        oasis_to_bids.convert_images(args.dataset_directory, args.bids_directory)
        oasis_to_bids.convert_clinical_data(args.clinical_data_directory, args.bids_directory)



# coding: utf8


def fix_join(path, *paths):
    # This workaround is used in pipelines like DWIPreprocessingUsingT1
    # In the workflow.connect part, you can use some function that are used as string, causing an import error
    import os
    return os.path.join(path, *paths)



# coding: utf8

"""
   Test to verify coding style dans clinica
"""

__author__ = "Mauricio Diaz"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Mauricio Diaz"]
__license__ = "See LICENSE.txt file"
__version__ = "0.2.0"
__maintainer__ = "Mauricio Diaz"
__email__ = "mauricio.diaz@inria.fr"
__status__ = "Development"

import pycodestyle
from os.path import dirname, abspath, join


def test_coding_style():
    """Test that we conform to PEP-8."""
    style = pycodestyle.StyleGuide(
        quiet=False,
        ignore=['E203', 'E121', 'E123', 'E126', 'E133',
                'E226', 'E241', 'E242', 'E704', 'W503',
                'E501', 'W504', 'W505', 'W605'])
    clinica_folder = join(dirname(dirname(abspath(__file__))), 'clinica')
    result = style.check_files([clinica_folder])
    result.print_statistics()
    assert result.total_errors == 0, "Found code style errors (and warnings)."



# coding: utf8

__author__ = "Alexis Guyot"
__copyright__ = "Copyright 2016-2019, The Aramis Lab Team"
__credits__ = ["Alexis Guyot"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Alexis Guyot"
__email__ = "alexis.guyot@icm-institute.org"
__status__ = "Development"

import clinica.pipelines.engine as cpe


class T1FreeSurferTemplate(cpe.Pipeline):
    """FreeSurfer Longitudinal template class

    Creates a pipelines that runs the FreeSurfer longitudinal template
    initialisation module for each subjects in a .tsv-defined list of
    subjects/sessions. This requires a prior-run of
    t1-freesurfer on the .tsv.

    Todos: N/A

    Returns:
        A clinica pipeline object containing the T1FreeSurferTemplate pipeline.

    Raises:

    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        pass

    def check_custom_dependencies(self):
        """Check dependencies that cannot be listed in `info.json`
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """

        return ['subject_list',
                'session_list2',
                'caps_target_list',
                'unpcssd_sublist',
                'pcssd_capstargetlist',
                'overwrite_tsv']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """

        return []

    def build_input_node(self):
        """Build and connect an input node to the pipeline.

        This node is supposedly used to load BIDS inputs when this
        pipeline is not already connected to the output of a previous
        Clinica pipeline.  For the purpose of the example, we simply
        read input arguments given by the command line interface and
        transmitted here through the `self.parameters` dictionary and
        pass it to the `self.input_node` to further by used as input of
        the core nodes.
        """
        import os
        import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils
#        import t1_freesurfer_template_utils as utils
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        # check if cross-sectional pipeline run on all subjects
        checkinput_node_name = '0_check_input'
        checkinput_node = npe.Node(name=checkinput_node_name,
                                   interface=nutil.Function(
                                       input_names=['in_caps_dir',
                                                    'in_subject_list',
                                                    'in_session_list',
                                                    'in_overwrite_caps',
                                                    'in_working_directory',
                                                    'in_n_procs',
                                                    'in_rundir'],
                                       output_names=['out_subject_list',
                                                     'out_session_list2',
                                                     'out_caps_target_list',
                                                     'out_unpcssd_sublist',
                                                     'out_pcssd_capstargetlist',
                                                     'out_overwrite_tsv'],
                                       function=utils.process_input_node))
        checkinput_node.inputs.in_caps_dir = self.caps_directory
        checkinput_node.inputs.in_subject_list = self.subjects
        checkinput_node.inputs.in_session_list = self.sessions
        checkinput_node.inputs.in_working_directory = self.base_dir
        checkinput_node.inputs.in_overwrite_caps = self.parameters[
            'overwrite_caps']
        checkinput_node.inputs.in_n_procs = self.parameters['n_procs']
        checkinput_node.inputs.in_rundir = os.getcwd()

        self.connect([
            (
                checkinput_node, self.input_node,
                [('out_subject_list', 'subject_list'),
                 ('out_session_list2', 'session_list2'),
                 ('out_caps_target_list', 'caps_target_list'),
                 ('out_unpcssd_sublist', 'unpcssd_sublist'),
                 ('out_pcssd_capstargetlist', 'pcssd_capstargetlist'),
                 ('out_overwrite_tsv', 'overwrite_tsv')])
        ])

    def build_output_node(self):
        """copy the template folder to CAPS DIR

        Will take all the data stored into the node in working directory
        that contains the built templates and copy those to the CAPS
        directory
        """
        pass

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline.
        """

        import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as utils
#        import t1_freesurfer_template_utils as utils
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        # Per-subject creation of the template

        # Step 1: create folder to store recon-all -base results
        # ======
        store_reconallbase_node_name = '1_store_reconallbase'
        store_reconallbase_node = npe.Node(
            name=store_reconallbase_node_name,
            interface=nutil.Function(
                input_names=['in_subject_list'],
                output_names=['out_workdirstore_path'],
                function=utils.store_reconallbase_results))

        # Step 2: prepare unbiased template flags for all subjects
        #         (i.e. create flags for recon-all -base)
        # ======
        prepare_template_node_name = '2_prepare_template'
        prepare_template_node = npe.MapNode(
            name=prepare_template_node_name,
            interface=nutil.Function(
                input_names=['in_subject',
                             'in_session_list'],
                output_names=['out_reconallbase_flags'],
                function=utils.get_reconallbase_flags
                ),
            iterfield=['in_subject',
                       'in_session_list'])

        # Step 3: create unbiased template for all subjects (i.e. run
        #         recon-all -base)
        # ======
        create_template_node_name = '3_create_template'
        create_template_node = npe.MapNode(
            name=create_template_node_name,
            interface=nutil.Function(
                input_names=['in_caps_dir',
                             'in_subject',
                             'in_session_list',
                             'in_reconallbase_flags',
                             'in_workdirstore_path'],
                output_names=['out_template_created'],
                function=utils.run_reconallbase
                ),
            iterfield=['in_subject',
                       'in_session_list',
                       'in_reconallbase_flags'])
        create_template_node.inputs.in_caps_dir = self.caps_directory

        # define the node to copy data to CAPS folder
        copy2caps_node_name = '4_copy2caps'
        copy2caps_node = npe.MapNode(
            name=copy2caps_node_name,
            interface=nutil.Function(
                input_names=['in_subject',
                             'in_subject_dir',
                             'in_caps_target',
                             'in_caps_dir',
                             'in_overwrite_caps',
                             'in_template_created'],
                output_names=['out_copy2_caps'],
                function=utils.copy_to_caps
                ),
            iterfield=['in_subject',
                       'in_caps_target'])
        copy2caps_node.inputs.in_caps_dir = self.caps_directory
        copy2caps_node.inputs.in_overwrite_caps = self.parameters[
            'overwrite_caps']

        # send data to the longitudinal correction pipeline (subjects
        # that have been detected as processed/non-processed and
        # corresponding sessions and CAPS locations)
        sendto_longcorr_node_name = '5_sendto_longcorr'
        sendto_longcorr_node = npe.Node(
            name=sendto_longcorr_node_name,
            interface=nutil.Function(
                input_names=['in_copy2_caps',
                             'in_unpcssd_sublist',
                             'in_pcssd_capstargetlist',
                             'in_overwrite_tsv'],
                output_names=['out_unpcssd_sublist',
                              'out_pcssd_capstargetlist',
                              'out_overwrite_tsv'],
                function=utils.sendto_longcorr))

        # Connection
        # ==========
        self.connect([
            # store_reconallbase_node
            (
                self.input_node, store_reconallbase_node,
                [('subject_list', 'in_subject_list')]),
            # prepare_template_node
            (
                self.input_node, prepare_template_node,
                [('subject_list', 'in_subject'),
                 ('session_list2', 'in_session_list')]),
            # create_template_node
            (
                self.input_node, create_template_node,
                [('subject_list', 'in_subject'),
                 ('session_list2', 'in_session_list')]),
            (
                prepare_template_node, create_template_node,
                [('out_reconallbase_flags', 'in_reconallbase_flags')]),
            (
                store_reconallbase_node, create_template_node,
                [('out_workdirstore_path', 'in_workdirstore_path')]),
            # copy2caps_node
            (
                self.input_node, copy2caps_node,
                [('subject_list', 'in_subject'),
                 ('caps_target_list', 'in_caps_target')]),
            (
                store_reconallbase_node, copy2caps_node,
                [('out_workdirstore_path', 'in_subject_dir')]),
            (
                create_template_node, copy2caps_node,
                [('out_template_created', 'in_template_created')]),
            # self.output_node (copy2caps_node)
            # data for sendto_longcorr_node
            (
                self.input_node, sendto_longcorr_node,
                [('unpcssd_sublist', 'in_unpcssd_sublist'),
                 ('pcssd_capstargetlist', 'in_pcssd_capstargetlist'),
                 ('overwrite_tsv', 'in_overwrite_tsv')]),
            (
                copy2caps_node, sendto_longcorr_node,
                [('out_copy2_caps', 'in_copy2_caps')])
        ])

        return self



# coding: utf8

import clinica.engine as ce


class StatisticsSurfaceCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 'statistics-surface'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Surface-based mass-univariate analysis with SurfStat:\n'
                             'http://clinica.run/doc/Pipelines/Stats_Surface/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        from colorama import Fore
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group('%sMandatory arguments%s' % (Fore.BLUE, Fore.RESET))
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("subject_visits_with_covariates_tsv",
                                  help='TSV file containing a list of subjects with their sessions and all '
                                       'the covariates and factors needed for the GLM.')
        clinica_comp.add_argument("design_matrix",
                                  help='String to define the design matrix that fits into the GLM, '
                                       'e.g. 1 + group + sex + age.')
        clinica_comp.add_argument("contrast",
                                  help='String to define the contrast matrix for the GLM, e.g. group. Please note '
                                       'that, when you want to perform negative correlation, the sign is ignored '
                                       'by the command line.')
        clinica_comp.add_argument("string_format",
                                  help='String to define the format of the columns in the TSV file, e.g., '
                                       '%%s %%s %%s %%f if the columns contain a string, a string, a string '
                                       'and a number, respectively.')
        clinica_comp.add_argument("group_id",
                                  help='User-defined identifier for the provided group of subjects.')
        clinica_comp.add_argument("glm_type",
                                  help='String based on the GLM type for the hypothesis. You can choose '
                                       'between group_comparison and correlation.')
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-fwhm", "--full_width_at_half_maximum",
                              type=int, default=20,
                              help='FWHM for the surface smoothing '
                                   '(default: --full_width_at_half_maximum %(default)s).')
        optional.add_argument("-ft", "--feature_type",
                              type=str, default='cortical_thickness',
                              help='Type of surface-based feature: cortical_thickness (from t1-freesurfer pipeline)'
                                   'or pet_fdg_projection (from pet-surface pipeline)'
                                   '(default: --feature_type %(default)s).')
        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments(add_tsv_flag=False)
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-cf", "--custom_file",
                              type=str, default=None,
                              help='Pattern of file inside CAPS directory using @subject, @session, @fwhm, @hemi. '
                                   'No --feature_type must be specified in order to use this flag. '
                                   'If you use this flag, you must specify a label with the --feature_label flag). '
                                   'See Wiki for an example.')
        advanced.add_argument("-fl", "--feature_label",
                              type=str, default=None,
                              help='Name of the feature type, it will be saved on the CAPS _measure-FEATURE_LABEL '
                                   'key-value association.')
        advanced.add_argument("-tup", "--threshold_uncorrected_pvalue",
                              type=float, default=0.001,
                              help='Threshold to display the uncorrected p-value '
                                   '(--threshold_uncorrected_pvalue %(default)s).')
        advanced.add_argument("-tcp", "--threshold_corrected_pvalue",
                              type=float, default=0.05,
                              help='Threshold to display the corrected p-value '
                                   '(default: --threshold_corrected_pvalue %(default)s)')
        advanced.add_argument("-ct", "--cluster_threshold",
                              type=float, default=0.001,
                              help='Threshold to define a cluster in the process of cluster-wise correction '
                                   '(default: --cluster_threshold %(default)s).')

    def run_command(self, args):
        """Run the pipeline with defined args."""
        import os
        from networkx import Graph
        from .statistics_surface_pipeline import StatisticsSurface
        from .statistics_surface_utils import (check_inputs,
                                               get_t1_freesurfer_custom_file,
                                               get_fdg_pet_surface_custom_file)
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit
        from clinica.utils.exceptions import ClinicaException

        if args.feature_type is not None:
            if args.custom_file is not None:
                raise ClinicaException('--feature_type and --custom_file are mutually exclusive: you must choose '
                                       'between one or the other. See documentation for more information.')
            if args.feature_label is not None:
                raise ClinicaException('--feature_label should not be used with --feature_type.')
            # FreeSurfer cortical thickness
            if args.feature_type == 'cortical_thickness':
                args.custom_file = get_t1_freesurfer_custom_file()
                args.feature_label = 'ct'
            # PET cortical projection
            elif args.feature_type == 'pet_fdg_projection':
                args.custom_file = get_fdg_pet_surface_custom_file()
                args.feature_label = 'fdg'
            else:
                raise ClinicaException('Feature type ' + args.feature_type + ' not recognized. Use --custom_file '
                                       'to specify your own files (without --feature_type).')
        elif args.feature_type is None:
            if args.custom_file is None:
                cprint('No feature type selected: using cortical thickness as default value')
                args.custom_file = get_t1_freesurfer_custom_file()
                args.feature_label = 'ct'
            else:
                cprint('Using custom features.')
                if args.feature_label is None:
                    raise ClinicaException('You must specify a --feature_label when using the --custom_files flag.')

        # Check if the group label has been existed, if yes, give an error to the users
        # Note(AR): if the user wants to compare Cortical Thickness measure with PET measure
        # using the group_id, Clinica won't allow it.
        # TODO: Modify this behaviour
        if os.path.exists(os.path.join(os.path.abspath(self.absolute_path(args.caps_directory)), 'groups', 'group-' + args.group_id)):
            error_message = 'group_id: ' + args.group_id + ' already exists, please choose another one or delete ' \
                            'the existing folder and also the working directory and rerun the pipeline'
            raise ClinicaException(error_message)
        parameters = {
            'group_label': args.group_id,
            'design_matrix': args.design_matrix,
            'contrast': args.contrast,
            'str_format': args.string_format,
            'glm_type': args.glm_type,
            'custom_file': args.custom_file,
            'feature_label': args.feature_label,
            'full_width_at_half_maximum': args.full_width_at_half_maximum,
            'threshold_uncorrected_pvalue': args.threshold_uncorrected_pvalue,
            'threshold_corrected_pvalue': args.threshold_corrected_pvalue,
            'cluster_threshold': args.cluster_threshold,
        }
        pipeline = StatisticsSurface(
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subject_visits_with_covariates_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        check_inputs(pipeline.caps_directory,
                     pipeline.parameters['custom_file'],
                     pipeline.parameters['full_width_at_half_maximum'],
                     pipeline.tsv_file)

        cprint("Parameters used for this pipeline:")
        cprint(pipeline.parameters)

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8

"""This module contains workflows for DWI preprocessing."""


def prepare_b0(num_b0s, low_bval=5, name='prepare_b0'):
    """
    Create a pipelines that prepare the data for further corrections. This
    pipelines coregister the B0 images and then
    average it in order to obtain only one average B0 images.
    The b-vectors and b-values are updated according to the modifications.

    Args:
        num_b0s (int): Number of b0 in the DWI dataset
        low_bval:
        name (Optional[str]): Name of the workflow

    Inputnode:
        in_dwi (str): Input DWI file.
        in_bvec (str): Vector file of the diffusion directions
            of the dwi dataset.
        in_bval (str): B-values file.

    Outputnode:
        dwi_b0_merge: Average of B0 images merged to the DWIs
        b0_reference: Average of the B0 images or the only B0 image
        out_bvec: Updated gradient vectors table
        out_bvals: Updated gradient values table
        mask_b0: Binary mask obtained from the average of the B0 images

    Returns:
        The workflow
    """
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe
    from nipype.workflows.dmri.fsl.utils import b0_average

    from clinica.utils.dwi import b0_dwi_split, insert_b0_into_dwi

    inputnode = pe.Node(interface=niu.IdentityInterface(
        fields=["in_dwi", "in_bvec", "in_bval"]),
        name="inputnode")

    b0_dwi_split = pe.Node(niu.Function(
        input_names=['in_file', 'in_bvals', 'in_bvecs', 'low_bval'],
        output_names=['out_b0', 'out_dwi', 'out_bvals', 'out_bvecs'],
        function=b0_dwi_split), name='b0_dwi_split')
    b0_dwi_split.inputs.lowbval = low_bval

    b0_flirt = b0_flirt_pipeline(num_b0s=num_b0s, name='b0_co_registration')

    b0_avg = pe.Node(niu.Function(
        input_names=['in_file'], output_names=['out_file'],
        function=b0_average), name='b0_average')

    mask_b0 = pe.Node(fsl.BET(frac=0.3, mask=True, robust=True),
                      name='mask_b0')

    insert_b0_into_dwi = pe.Node(niu.Function(
        input_names=['in_b0', 'in_dwi', 'in_bvals', 'in_bvecs'],
        output_names=['out_dwi', 'out_bvals', 'out_bvecs'],
        function=insert_b0_into_dwi), name='insert_b0avg_into_dwi')

    outputnode = pe.Node(niu.IdentityInterface(
        fields=['mask_b0', 'b0_reference', 'dwi_b0_merge',
                'out_bvecs', 'out_bvals']),
        name='outputnode')

    wf = pe.Workflow(name=name)

    if num_b0s == 1:
        wf.connect([
            # Split dataset into two datasets (b0, b>low_bval)
            (inputnode,          b0_dwi_split, [('in_bval',       'in_bvals'),
                                                ('in_bvec',       'in_bvecs'),
                                                ('in_dwi',       'in_file')]),
            # Merge datasets such that bval(DWI) = (0 b1 ... bn)
            (b0_dwi_split, insert_b0_into_dwi, [('out_b0',           'in_b0'),
                                                ('out_dwi',         'in_dwi'),
                                                ('out_bvals',     'in_bvals'),
                                                ('out_bvecs',   'in_bvecs')]),
            # Compute b0 mask
            (b0_dwi_split,            mask_b0, [('out_b0',       'in_file')]),
            # Outputnode
            (insert_b0_into_dwi,   outputnode, [('out_dwi',   'dwi_b0_merge'),
                                                ('out_bvals',    'out_bvals'),
                                                ('out_bvecs',  'out_bvecs')]),
            (mask_b0,              outputnode, [('mask_file',    'mask_b0')]),
            (b0_dwi_split,         outputnode, [('out_b0',  'b0_reference')])
        ])
    elif num_b0s > 1:
        wf.connect([
            # Split dataset into two datasets (b0s, b>low_bval)
            (inputnode,             b0_dwi_split, [('in_bval',              'in_bvals'),  # noqa
                                                   ('in_bvec',              'in_bvecs'),  # noqa
                                                   ('in_dwi',              'in_file')]),  # noqa
            # Register the b0 onto the first b0
            (b0_dwi_split,              b0_flirt, [('out_b0',    'inputnode.in_file')]),  # noqa
            # Average the b0s
            (b0_flirt,                    b0_avg, [('outputnode.out_file', 'in_file')]),  # noqa
            # Compute b0 mask from b0avg
            (b0_avg,                     mask_b0, [('out_file',            'in_file')]),  # noqa
            # Merge datasets such that bval(DWI) = (0 b1 ... bn)
            (b0_avg,          insert_b0_into_dwi, [('out_file',              'in_b0')]),  # noqa
            (b0_dwi_split,    insert_b0_into_dwi, [('out_dwi',                'in_dwi'),  # noqa
                                                   ('out_bvals',            'in_bvals'),  # noqa
                                                   ('out_bvecs',          'in_bvecs')]),  # noqa
            # Outputnode
            (insert_b0_into_dwi,      outputnode, [('out_dwi',          'dwi_b0_merge'),  # noqa
                                                   ('out_bvals',           'out_bvals'),  # noqa
                                                   ('out_bvecs',         'out_bvecs')]),  # noqa
            (mask_b0,                 outputnode, [('mask_file',           'mask_b0')]),  # noqa
            (b0_avg,                  outputnode, [('out_file',       'b0_reference')])   # noqa
        ])
    else:
        raise ValueError('The number of b0s should be strictly positive.')

    return wf


def b0_flirt_pipeline(num_b0s, name='b0_coregistration'):
    """
    Rigid registration of the B0 dataset onto the first volume. Rigid
    registration is achieved using FLIRT and the normalized
    correlation.

    Args:
        num_b0s (int): Number of the B0 volumes in the dataset.
        name (str): Name of the workflow.

    Inputnode:
        in_file(str): B0 dataset.

    Outputnode
        out_b0_reg(str): The set of B0 volumes registered to the first volume.

    Returns:
        The workflow
    """
    import nipype.pipeline.engine as pe
    from nipype.interfaces import fsl
    import nipype.interfaces.utility as niu

    from clinica.utils.dwi import merge_volumes_tdim

    inputnode = pe.Node(niu.IdentityInterface(fields=['in_file']),
                        name='inputnode')
    fslroi_ref = pe.Node(fsl.ExtractROI(args='0 1'), name='b0_reference')
    tsize = num_b0s - 1
    fslroi_moving = pe.Node(fsl.ExtractROI(args='1 '+str(tsize)),
                            name='b0_moving')
    split_moving = pe.Node(fsl.Split(dimension='t'), name='split_b0_moving')

    bet_ref = pe.Node(fsl.BET(frac=0.3, mask=True, robust=True),
                      name='bet_ref')

    dilate = pe.Node(
            fsl.maths.MathsCommand(
                nan2zeros=True,
                args='-kernel sphere 5 -dilM'),
            name='mask_dilate')

    flirt = pe.MapNode(fsl.FLIRT(
        interp='spline', dof=6, bins=50, save_log=True,
        cost='corratio', cost_func='corratio', padding_size=10,
        searchr_x=[-4, 4], searchr_y=[-4, 4], searchr_z=[-4, 4],
        fine_search=1, coarse_search=10),
        name='b0_co_registration', iterfield=['in_file'])

    merge = pe.Node(fsl.Merge(dimension='t'), name='merge_registered_b0s')
    thres = pe.MapNode(fsl.Threshold(thresh=0.0), iterfield=['in_file'],
                       name='remove_negative')
    insert_ref = pe.Node(niu.Function(input_names=['in_file1', 'in_file2'],
                                      output_names=['out_file'],
                                      function=merge_volumes_tdim),
                         name='concat_ref_moving')

    outputnode = pe.Node(niu.IdentityInterface(
        fields=['out_file', 'out_xfms']),
        name='outputnode')

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode,  fslroi_ref,   [('in_file', 'in_file')]),
        (inputnode,  fslroi_moving,   [('in_file', 'in_file')]),
        (fslroi_moving, split_moving,   [('roi_file', 'in_file')]),
        (fslroi_ref, bet_ref, [('roi_file', 'in_file')]),
        (bet_ref, dilate, [('mask_file', 'in_file')]),
        (dilate, flirt, [('out_file', 'ref_weight'),
                         ('out_file', 'in_weight')]),
        (fslroi_ref, flirt, [('roi_file', 'reference')]),
        (split_moving, flirt, [('out_files', 'in_file')]),
        (flirt, thres, [('out_file', 'in_file')]),
        (thres, merge, [('out_file', 'in_files')]),
        (merge, insert_ref, [('merged_file', 'in_file2')]),
        (fslroi_ref, insert_ref, [('roi_file', 'in_file1')]),
        (insert_ref, outputnode, [('out_file', 'out_file')]),
        (flirt, outputnode, [('out_matrix_file', 'out_xfms')])
    ])
    return wf


def dwi_flirt(name='DWICoregistration', excl_nodiff=False, flirt_param={}):
    """
    Generates a workflow for linear registration of dwi volumes using flirt.

    Inputnode
    ---------
    reference : FILE
      Mandatory input. Reference data set.
    in_file : FILE
      Mandatory input. Moving data set.
    ref_mask : FILE
      Mandatory input. Binary mask of the reference volume.
    in_xfms : FILE
      Mandatory input. Intialisation matrices for flirt.
    in_bval : FILE
      Mandatory input. B values file.

    """
    import nipype.interfaces.ants as ants
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe

    from nipype.workflows.dmri.fsl.utils import _checkinitxfm

    from nipype.workflows.dmri.fsl.utils import enhance

    inputnode = pe.Node(
            niu.IdentityInterface(
                fields=['reference',
                        'in_file',
                        'ref_mask',
                        'in_xfms',
                        'in_bval']),
            name='inputnode')

    initmat = pe.Node(
            niu.Function(
                input_names=['in_bval',
                             'in_xfms',
                             'excl_nodiff'],
                output_names=['init_xfms'],
                function=_checkinitxfm),
            name='InitXforms')
    initmat.inputs.excl_nodiff = excl_nodiff
    dilate = pe.Node(
            fsl.maths.MathsCommand(
                nan2zeros=True,
                args='-kernel sphere 5 -dilM'),
            name='MskDilate')
    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    n4 = pe.Node(ants.N4BiasFieldCorrection(dimension=3), name='Bias')
    flirt = pe.MapNode(fsl.FLIRT(**flirt_param), name='CoRegistration',
                       iterfield=['in_file', 'in_matrix_file'])
    thres = pe.MapNode(fsl.Threshold(thresh=0.0), iterfield=['in_file'],
                       name='RemoveNegative')
    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeDWIs')
    outputnode = pe.Node(
            niu.IdentityInterface(
                fields=['out_file',
                        'out_xfms',
                        'out_ref']),
            name='outputnode')
    enhb0 = pe.Node(niu.Function(
        input_names=['in_file', 'in_mask', 'clip_limit'],
        output_names=['out_file'], function=enhance), name='B0Equalize')
    enhb0.inputs.clip_limit = 0.015
    enhdw = pe.MapNode(niu.Function(
        input_names=['in_file', 'in_mask'], output_names=['out_file'],
        function=enhance), name='DWEqualize', iterfield=['in_file'])
    # enhb0.inputs.clip_limit = clip_limit

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode,  split,      [('in_file', 'in_file')]),
        (inputnode,  dilate,     [('ref_mask', 'in_file')]),
        (inputnode,   n4,        [('reference', 'input_image'),
                                  ('ref_mask', 'mask_image')]),
        #        (inputnode,  flirt,      [('ref_mask', 'reference')]),
        (n4, enhb0, [('output_image', 'in_file')]),
        (enhb0, flirt, [('out_file', 'reference')]),
        (inputnode, initmat, [('in_xfms', 'in_xfms'),
                              ('in_bval', 'in_bval')]),
        (split, enhdw, [('out_files', 'in_file')]),
        (dilate, enhdw, [('out_file', 'in_mask')]),
        (dilate, flirt, [('out_file', 'ref_weight'),
                         ('out_file', 'in_weight')]),
        (enhdw, flirt, [('out_file', 'in_file')]),
        (initmat, flirt, [('init_xfms', 'in_matrix_file')]),
        (flirt,      thres,      [('out_file', 'in_file')]),
        (thres,      merge,      [('out_file', 'in_files')]),
        (merge,     outputnode, [('merged_file', 'out_file')]),
        (enhb0, outputnode, [('out_file', 'out_ref')]),
        (flirt,     outputnode, [('out_matrix_file', 'out_xfms')])
    ])
    return wf


def hmc_pipeline(name='motion_correct'):
    """
    HMC stands for head-motion correction.

    Creates a pipelines that corrects for head motion artifacts in dMRI
    sequences. It takes a series of diffusion weighted images and
    rigidly co-registers them to one reference image (FLIRT normalised
    mutual information). Finally, the `b`-matrix is rotated
    accordingly [Leemans09]_ making use of the rotation matrix
    obtained by FLIRT.

    A list of rigid transformation matrices is provided, so that transforms
    can be chained.
    This is useful to correct for artifacts with only one interpolation process
    and also to compute nuisance regressors as proposed by [Yendiki13]_.

    .. warning:: This workflow rotates the `b`-vectors, so please be advised
      that not all the dicom converters ensure the consistency between the
      resulting nifti orientation and the gradients table (e.g. dcm2nii
      checks it).

    .. admonition:: References

      .. [Leemans09] Leemans A, and Jones DK, `The B-matrix must be rotated
        when correcting for subject motion in DTI data
        <http://dx.doi.org/10.1002/mrm.21890>`_,
        Magn Reson Med. 61(6):1336-49. 2009. doi: 10.1002/mrm.21890.

      .. [Yendiki13] Yendiki A et al., `Spurious group differences due to head
        motion in a diffusion MRI study
        <http://dx.doi.org/10.1016/j.neuroimage.2013.11.027>`_.
        Neuroimage. 21(88C):79-90. 2013. doi: 10.1016/j.neuroimage.2013.11.027

    Inputnode
    ---------
    in_file : FILE
      Mandatory input. Input dwi file.
    in_bvec : FILE
      Mandatory input. Vector file of the diffusion directions of the dwi dataset.
    in_bval : FILE
      Mandatory input. B values file.
    in_mask : FILE
      Mandatory input. Weights mask of reference image (a file with data
      range in [0.0, 1.0], indicating the weight of each voxel when computing the metric
    ref_num : INT
      Optional input. Default=0. Index of the b0 volume that should be taken as reference.

    Outputnode
    ----------
        outputnode.out_file - corrected dwi file
        outputnode.out_bvec - rotated gradient vectors table
        outputnode.out_xfms - list of transformation matrices

    """
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe

    from nipype.workflows.data import get_flirt_schedule
    from nipype.workflows.dmri.fsl.utils import insert_mat
    from nipype.workflows.dmri.fsl.utils import rotate_bvecs

    from clinica.utils.dwi import merge_volumes_tdim
    from clinica.utils.dwi import hmc_split
    from clinica.workflows.dwi_preprocessing import dwi_flirt

    params = dict(dof=6, bgvalue=0, save_log=True, no_search=True,
                  # cost='mutualinfo', cost_func='mutualinfo', bins=64,
                  schedule=get_flirt_schedule('hmc'))

    inputnode = pe.Node(niu.IdentityInterface(
        fields=['in_file', 'in_bvec', 'in_bval', 'in_mask', 'ref_num']),
        name='inputnode')

    split = pe.Node(
            niu.Function(
                function=hmc_split,
                input_names=['in_file', 'in_bval', 'ref_num'],
                output_names=['out_ref', 'out_mov', 'out_bval', 'volid']),
            name='split_ref_moving')

    flirt = dwi_flirt(flirt_param=params)

    insmat = pe.Node(niu.Function(
        input_names=['inlist', 'volid'],
        output_names=['out'],
        function=insert_mat), name='insert_ref_matrix')

    rot_bvec = pe.Node(niu.Function(
        input_names=['in_bvec', 'in_matrix'],
        output_names=['out_file'],
        function=rotate_bvecs), name='Rotate_Bvec')

    merged_volumes = pe.Node(niu.Function(
        input_names=['in_file1', 'in_file2'],
        output_names=['out_file'],
        function=merge_volumes_tdim), name='merge_reference_moving')

    outputnode = pe.Node(niu.IdentityInterface(
        fields=['out_file', 'out_bvec', 'out_xfms', 'mask_B0']),
        name='outputnode')

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode,          split,  [('in_file',                'in_file'),
                                      ('in_bval',                'in_bval'),
                                      ('ref_num',              'ref_num')]),
        (inputnode,          flirt,  [('in_mask',   'inputnode.ref_mask')]),
        (split,              flirt,  [('out_ref',    'inputnode.reference'),
                                      ('out_mov',      'inputnode.in_file'),
                                      ('out_bval',   'inputnode.in_bval')]),
        (flirt,              insmat, [('outputnode.out_xfms',   'inlist')]),
        (split,              insmat, [('volid',                  'volid')]),
        (inputnode,        rot_bvec, [('in_bvec',              'in_bvec')]),
        (insmat,           rot_bvec, [('out',                'in_matrix')]),
        (rot_bvec,       outputnode, [('out_file',            'out_bvec')]),
        (flirt,      merged_volumes, [('outputnode.out_ref',    'in_file1'),
                                      ('outputnode.out_file', 'in_file2')]),
        (merged_volumes, outputnode, [('out_file',            'out_file')]),
        (insmat,         outputnode, [('out',                 'out_xfms')]),
        (flirt,          outputnode, [('outputnode.out_ref',   'mask_B0')])
    ])
    return wf


def ecc_pipeline(name='eddy_correct'):
    """
    ECC stands for Eddy currents correction.
    Creates a pipelines that corrects for artifacts induced by Eddy currents in
    dMRI sequences.
    It takes a series of diffusion weighted images and linearly co-registers
    them to one reference image (the average of all b0s in the dataset).
    DWIs are also modulated by the determinant of the Jacobian as indicated by
    [Jones10]_ and [Rohde04]_.
    A list of rigid transformation matrices can be provided, sourcing from a
    :func:`.hmc_pipeline` workflow, to initialize registrations in a *motion
    free* framework.
    A list of affine transformation matrices is available as output, so that
    transforms can be chained (discussion
    `here <https://github.com/nipy/nipype/pull/530#issuecomment-14505042>`_).
    .. admonition:: References
      .. [Jones10] Jones DK, `The signal intensity must be modulated by the
        determinant of the Jacobian when correcting for eddy currents in
        diffusion MRI
        <http://cds.ismrm.org/protected/10MProceedings/files/1644_129.pdf>`_,
        Proc. ISMRM 18th Annual Meeting, (2010).
      .. [Rohde04] Rohde et al., `Comprehensive Approach for Correction of
        Motion and Distortion in Diffusion-Weighted MRI
        <http://stbb.nichd.nih.gov/pdf/com_app_cor_mri04.pdf>`_, MRM
        51:103-114 (2004).
    Example
    -------
    from nipype.workflows.dmri.fsl.artifacts import ecc_pipeline
    ecc = ecc_pipeline()
    ecc.inputs.inputnode.in_file = 'diffusion.nii'
    ecc.inputs.inputnode.in_bval = 'diffusion.bval'
    ecc.inputs.inputnode.in_mask = 'mask.nii'
    ecc.run() # doctest: +SKIP
    Inputs::
        inputnode.in_file - input dwi file
        inputnode.in_mask - weights mask of reference image (a file with data \
range sin [0.0, 1.0], indicating the weight of each voxel when computing the \
metric.
        inputnode.in_bval - b-values table
        inputnode.in_xfms - list of matrices to initialize registration (from \
head-motion correction)
    Outputs::
        outputnode.out_file - corrected dwi file
        outputnode.out_xfms - list of transformation matrices
    """

    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe
    import nipype.interfaces.fsl as fsl

    from nipype.workflows.data import get_flirt_schedule
    from nipype.workflows.dmri.fsl.utils import extract_bval
    from nipype.workflows.dmri.fsl.utils import recompose_xfm
    from nipype.workflows.dmri.fsl.utils import recompose_dwi
    from nipype.workflows.dmri.fsl.artifacts import _xfm_jacobian

    from clinica.workflows.dwi_preprocessing import dwi_flirt
    from clinica.utils.dwi import merge_volumes_tdim

    params = dict(dof=12, no_search=True, interp='spline', bgvalue=0,
                  schedule=get_flirt_schedule('ecc'))

    inputnode = pe.Node(niu.IdentityInterface(
        fields=['in_file', 'in_bval', 'in_mask', 'in_xfms']), name='inputnode')

    getb0 = pe.Node(fsl.ExtractROI(t_min=0, t_size=1), name='get_b0')

    pick_dws = pe.Node(niu.Function(
        input_names=['in_dwi', 'in_bval', 'b'], output_names=['out_file'],
        function=extract_bval), name='extract_dwi')
    pick_dws.inputs.b = 'diff'

    flirt = dwi_flirt(flirt_param=params, excl_nodiff=True)

    mult = pe.MapNode(fsl.BinaryMaths(operation='mul'), name='ModulateDWIs',
                      iterfield=['in_file', 'operand_value'])
    thres = pe.MapNode(fsl.Threshold(thresh=0.0), iterfield=['in_file'],
                       name='RemoveNegative')

    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    get_mat = pe.Node(niu.Function(
        input_names=['in_bval', 'in_xfms'], output_names=['out_files'],
        function=recompose_xfm), name='GatherMatrices')
    merge = pe.Node(niu.Function(
        input_names=['in_dwi', 'in_bval', 'in_corrected'],
        output_names=['out_file'], function=recompose_dwi), name='MergeDWIs')

    merged_volumes = pe.Node(niu.Function(
        input_names=['in_file1', 'in_file2'],
        output_names=['out_file'],
        function=merge_volumes_tdim), name='merge_enhanced_ref_dwis')

    outputnode = pe.Node(niu.IdentityInterface(
        fields=['out_file', 'out_xfms']), name='outputnode')

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode,  getb0,        [('in_file', 'in_file')]),
        (inputnode,  pick_dws,     [('in_file', 'in_dwi'),
                                    ('in_bval', 'in_bval')]),
        (flirt, merged_volumes, [('outputnode.out_ref', 'in_file1'),
                                 ('outputnode.out_file', 'in_file2')]),

        (merged_volumes,  merge,        [('out_file', 'in_dwi')]),
        (inputnode, merge, [('in_bval', 'in_bval')]),
        (inputnode,  flirt,        [('in_mask', 'inputnode.ref_mask'),
                                    ('in_xfms', 'inputnode.in_xfms'),
                                    ('in_bval', 'inputnode.in_bval')]),
        (inputnode,  get_mat,      [('in_bval', 'in_bval')]),
        (getb0,      flirt,        [('roi_file', 'inputnode.reference')]),
        (pick_dws,   flirt,        [('out_file', 'inputnode.in_file')]),
        (flirt,      get_mat,      [('outputnode.out_xfms', 'in_xfms')]),
        (flirt,      mult,         [(('outputnode.out_xfms', _xfm_jacobian),
                                     'operand_value')]),
        (flirt,      split,        [('outputnode.out_file', 'in_file')]),
        (split,      mult,         [('out_files', 'in_file')]),
        (mult,       thres,        [('out_file', 'in_file')]),
        (thres,      merge,        [('out_file', 'in_corrected')]),
        (get_mat,    outputnode,   [('out_files', 'out_xfms')]),
        (merge,      outputnode,   [('out_file', 'out_file')])
    ])
    return wf


def remove_bias(name='bias_correct'):
    """
    This workflow estimates a single multiplicative bias field from the
    averaged *b0* image, as suggested in [Jeurissen2014]_.
    .. admonition:: References
      .. [Jeurissen2014] Jeurissen B. et al., `Multi-tissue constrained
        spherical deconvolution for improved analysis of multi-shell diffusion
        MRI data <http://dx.doi.org/10.1016/j.neuroimage.2014.07.061>`_.squeue

        NeuroImage (2014). doi: 10.1016/j.neuroimage.2014.07.061
    Example
    -------
    >>> from nipype.workflows.dmri.fsl.artifacts import remove_bias
    >>> bias = remove_bias()
    >>> bias.inputs.inputnode.in_file = 'epi.nii'
    >>> bias.inputs.inputnode.in_bval = 'diffusion.bval'
    >>> bias.inputs.inputnode.in_mask = 'mask.nii'
    >>> bias.run() # doctest: +SKIP
    """
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.ants as ants

    inputnode = pe.Node(niu.IdentityInterface(
        fields=['in_file']), name='inputnode')

    outputnode = pe.Node(niu.IdentityInterface(fields=['out_file', 'b0_mask']),
                         name='outputnode')

    get_b0 = pe.Node(fsl.ExtractROI(t_min=0, t_size=1), name='get_b0')

    mask_b0 = pe.Node(fsl.BET(frac=0.3, mask=True, robust=True),
                      name='mask_b0')

    n4 = pe.Node(ants.N4BiasFieldCorrection(
        dimension=3, save_bias=True, bspline_fitting_distance=600),
        name='Bias_b0')
    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    mult = pe.MapNode(fsl.MultiImageMaths(op_string='-div %s'),
                      iterfield=['in_file'], name='RemoveBiasOfDWIs')
    thres = pe.MapNode(fsl.Threshold(thresh=0.0), iterfield=['in_file'],
                       name='RemoveNegative')
    merge = pe.Node(fsl.utils.Merge(dimension='t'), name='MergeDWIs')

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, get_b0, [('in_file', 'in_file')]),
        (get_b0,   n4, [('roi_file', 'input_image')]),
        (get_b0, mask_b0, [('roi_file', 'in_file')]),
        (mask_b0, n4, [('mask_file', 'mask_image')]),
        (inputnode, split, [('in_file', 'in_file')]),
        (n4,    mult, [('bias_image', 'operand_files')]),
        (split, mult, [('out_files', 'in_file')]),
        (mult, thres, [('out_file', 'in_file')]),
        (thres, merge, [('out_file', 'in_files')]),
        (merge,   outputnode, [('merged_file', 'out_file')]),
        (mask_b0, outputnode, [('mask_file', 'b0_mask')])
    ])
    return wf


def eddy_fsl_pipeline(epi_param, name='eddy_fsl'):
    """
    Using eddy from fsl for head motion correction and eddy current distortion correction.

    """
    from nipype.interfaces.fsl import Eddy
    import nipype.interfaces.utility as niu     # utility
    import nipype.pipeline.engine as pe          # pypeline engine
    from clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_utils import eddy_fsl, generate_acq, generate_index, b0_indices

    inputnode = pe.Node(
            niu.IdentityInterface(
                fields=['in_file',
                        'in_bvec',
                        'in_bval',
                        'in_mask',
                        'ref_b0']),
            name='inputnode')

    generate_acq = pe.Node(niu.Function(function=generate_acq,
                                        input_names=['in_b0', 'epi_param'],
                                        output_names=['out_file']),
                           name='generate_acq')
    generate_acq.inputs.epi_param = epi_param

    list_b0 = pe.Node(niu.Function(input_names=['in_bval'],
                                   output_names=['out_idx'],
                                   function=b0_indices),
                      name='find_b0_indices')

    generate_index = pe.Node(niu.Function(function=generate_index,
                                          input_names=['in_bval', 'b0_index'],
                                          output_names=['eddy_index']),
                             name='generate_index')

    eddy = pe.Node(niu.Function(input_names=['in_bvec', 'in_bval', 'in_file', 'in_mask', 'in_acqp', 'in_index'],
                                output_names=['out_parameter', 'out_corrected', 'out_rotated_bvecs'],
                                function=eddy_fsl),
                   name='eddy_fsl')

    eddy = pe.Node(interface=Eddy(), name='eddy_fsl')
    eddy.inputs.flm = 'linear'

    outputnode = pe.Node(niu.IdentityInterface(fields=['out_parameter',
                                                       'out_corrected',
                                                       'out_rotated_bvecs']),
                         name='outputnode')

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode,     generate_acq,   [('ref_b0', 'in_b0')]),
        (inputnode,  generate_index,      [('in_bval', 'in_bval')]),
        (list_b0,  generate_index,      [('out_idx', 'b0_index')]),
        (inputnode,      list_b0,      [('in_bval', 'in_bval')]),

        (inputnode,      eddy,     [('in_bvec', 'in_bvec')]),
        (inputnode,      eddy,     [('in_bval', 'in_bval')]),
        (inputnode,  eddy,   [('in_file', 'in_file')]),
        (inputnode,     eddy,   [('in_mask', 'in_mask')]),
        (generate_acq,      eddy, [('out_file', 'in_acqp')]),
        (generate_index,      eddy, [('eddy_index', 'in_index')]),

        (eddy,   outputnode, [('out_parameter', 'out_parameter')]),
        (eddy,      outputnode, [('out_corrected', 'out_corrected')]),
        (eddy,      outputnode, [('out_rotated_bvecs', 'out_rotated_bvecs')])
    ])
    return wf


def epi_pipeline(name='susceptibility_distortion_correction_using_t1'):
    """
    This workflow allows to correct for echo-planareinduced susceptibility artifacts without fieldmap
    (e.g. ADNI Database) by elastically register DWIs to their respective baseline T1-weighted
    structural scans using an inverse consistent registration algorithm with a mutual information cost
    function (SyN algorithm). This workflow allows also a coregistration of DWIs with their respective
    baseline T1-weighted structural scans in order to latter combine tracks and cortex parcelation.
    ..  warning:: This workflow rotates the `b`-vectors'
    .. References
      .. Nir et al. (Neurobiology of Aging 2015)- Connectivity network measures predict volumetric atrophy in mild cognitive impairment

        Leow et al. (IEEE Trans Med Imaging 2007)- Statistical Properties of Jacobian Maps and the Realization of Unbiased Large Deformation Nonlinear Image Registration
    Example
    -------
    >>> epi = epi_pipeline()
    >>> epi.inputs.inputnode.DWI = 'DWI.nii'
    >>> epi.inputs.inputnode.bvec = 'bvec.txt'
    >>> epi.inputs.inputnode.T1 = 'T1.nii'
    >>> epi.run() # doctest: +SKIP
    """

    from clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_utils import create_jacobian_determinant_image, change_itk_transform_type, expend_matrix_list, rotate_bvecs, ants_registration_syn_quick, ants_warp_image_multi_transform, ants_combin_transform
    import nipype.pipeline.engine as pe
    import nipype.interfaces.utility as niu
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.c3 as c3

    inputnode = pe.Node(niu.IdentityInterface(fields=['T1', 'DWI', 'bvec']), name='inputnode')

    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    pick_ref = pe.Node(niu.Select(), name='Pick_b0')
    pick_ref.inputs.index = [0]

    flirt_b0_2_T1 = pe.Node(interface=fsl.FLIRT(dof=6), name='flirt_B0_2_T1')
    flirt_b0_2_T1.inputs.interp = "spline"
    flirt_b0_2_T1.inputs.cost = 'normmi'
    flirt_b0_2_T1.inputs.cost_func = 'normmi'

    apply_xfm = pe.Node(interface=fsl.preprocess.ApplyXFM(), name='apply_xfm')
    apply_xfm.inputs.apply_xfm = True

    expend_matrix = pe.Node(
            interface=niu.Function(
                input_names=['in_matrix', 'in_bvec'],
                output_names=['out_matrix_list'],
                function=expend_matrix_list),
            name='expend_matrix')

    rot_bvec = pe.Node(
            niu.Function(
                input_names=['in_matrix', 'in_bvec'],
                output_names=['out_file'],
                function=rotate_bvecs),
            name='Rotate_Bvec')

    antsRegistrationSyNQuick = pe.Node(
            interface=niu.Function(
                input_names=['fix_image', 'moving_image'],
                output_names=['image_warped',
                              'affine_matrix',
                              'warp',
                              'inverse_warped',
                              'inverse_warp'],
                function=ants_registration_syn_quick),
            name='antsRegistrationSyNQuick')

    c3d_flirt2ants = pe.Node(c3.C3dAffineTool(), name='fsl_reg_2_itk')
    c3d_flirt2ants.inputs.itk_transform = True
    c3d_flirt2ants.inputs.fsl2ras = True

    change_transform = pe.Node(niu.Function(
            input_names=['input_affine_file'],
            output_names=['updated_affine_file'],
            function=change_itk_transform_type),
            name='change_transform_type')

    merge_transform = pe.Node(niu.Merge(3), name='MergeTransforms')

    apply_transform = pe.MapNode(interface=niu.Function(input_names=['fix_image', 'moving_image', 'ants_warp_affine'],
                                                        output_names=['out_warp_field', 'out_warped'],
                                                        function=ants_combin_transform),
                                 iterfield=['moving_image'],
                                 name='warp_filed')

    jacobian = pe.MapNode(interface=niu.Function(input_names=['imageDimension', 'deformationField', 'outputImage'],
                                                 output_names=['outputImage'],
                                                 function=create_jacobian_determinant_image),
                          iterfield=['deformationField'],
                          name='jacobian')

    jacobian.inputs.imageDimension = 3
    jacobian.inputs.outputImage = 'Jacobian_image.nii.gz'

    jacmult = pe.MapNode(fsl.MultiImageMaths(op_string='-mul %s'),
                         iterfield=['in_file', 'operand_files'],
                         name='ModulateDWIs')

    thres = pe.MapNode(fsl.Threshold(thresh=0.0), iterfield=['in_file'],
                       name='RemoveNegative')

    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeDWIs')

    outputnode = pe.Node(niu.IdentityInterface(fields=['DWI_2_T1_Coregistration_matrix',
                                                       'epi_correction_deformation_field',
                                                       'epi_correction_affine_transform',
                                                       'epi_correction_image_warped',
                                                       'DWIs_epicorrected',
                                                       'warp_epi',
                                                       'out_bvec'
                                                       ]), name='outputnode')

    wf = pe.Workflow(name='epi_pipeline')

    wf.connect([(inputnode, split, [('DWI', 'in_file')])])
    wf.connect([(split, pick_ref, [('out_files', 'inlist')])])
    wf.connect([(pick_ref, flirt_b0_2_T1, [('out', 'in_file')])])
    wf.connect([(inputnode, flirt_b0_2_T1, [('T1', 'reference')])])
    wf.connect([(inputnode, rot_bvec, [('bvec', 'in_bvec')])])
    wf.connect([(flirt_b0_2_T1, expend_matrix, [('out_matrix_file', 'in_matrix')])])
    wf.connect([(inputnode, expend_matrix, [('bvec', 'in_bvec')])])
    wf.connect([(expend_matrix, rot_bvec, [('out_matrix_list', 'in_matrix')])])
    wf.connect([(inputnode, antsRegistrationSyNQuick, [('T1', 'fix_image')])])
    wf.connect([(flirt_b0_2_T1, antsRegistrationSyNQuick, [('out_file', 'moving_image')])])

    wf.connect([(inputnode, c3d_flirt2ants, [('T1', 'reference_file')])])
    wf.connect([(pick_ref, c3d_flirt2ants, [('out', 'source_file')])])
    wf.connect([(flirt_b0_2_T1, c3d_flirt2ants, [('out_matrix_file', 'transform_file')])])
    wf.connect([(c3d_flirt2ants, change_transform, [('itk_transform', 'input_affine_file')])])

    wf.connect([(antsRegistrationSyNQuick, merge_transform, [('warp', 'in1')])])
    wf.connect([(antsRegistrationSyNQuick, merge_transform, [('affine_matrix', 'in2')])])
    wf.connect([(change_transform, merge_transform, [('updated_affine_file', 'in3')])])
    wf.connect([(inputnode, apply_transform, [('T1', 'fix_image')])])
    wf.connect([(split, apply_transform, [('out_files', 'moving_image')])])

    wf.connect([(merge_transform, apply_transform, [('out', 'ants_warp_affine')])])
    wf.connect([(apply_transform, jacobian, [('out_warp_field', 'deformationField')])])
    wf.connect([(apply_transform, jacmult, [('out_warped', 'operand_files')])])
    wf.connect([(jacobian, jacmult, [('outputImage', 'in_file')])])
    wf.connect([(jacmult, thres, [('out_file', 'in_file')])])
    wf.connect([(thres, merge, [('out_file', 'in_files')])])

    wf.connect([(merge, outputnode, [('merged_file', 'DWIs_epicorrected')])])
    wf.connect([(flirt_b0_2_T1, outputnode, [('out_matrix_file', 'DWI_2_T1_Coregistration_matrix')])])
    wf.connect([(antsRegistrationSyNQuick, outputnode, [('warp', 'epi_correction_deformation_field'),
                                                        ('affine_matrix', 'epi_correction_affine_transform'),
                                                        ('image_warped', 'epi_correction_image_warped')])])
    wf.connect([(merge_transform, outputnode, [('out', 'warp_epi')])])
    wf.connect([(rot_bvec, outputnode, [('out_file', 'out_bvec')])])

    return wf



# coding: utf8


def ants_registration_syn_quick(fix_image, moving_image):
    import os
    import os.path as op

    image_warped = op.abspath('SyN_QuickWarped.nii.gz')
    affine_matrix = op.abspath('SyN_Quick0GenericAffine.mat')
    warp = op.abspath('SyN_Quick1Warp.nii.gz')
    inverse_warped = op.abspath('SyN_QuickInverseWarped.nii.gz')
    inverse_warp = op.abspath('SyN_Quick1InverseWarp.nii.gz')

    cmd = 'antsRegistrationSyNQuick.sh -t br -d 3 -f %s  -m %s -o SyN_Quick' \
          % (fix_image, moving_image)
    os.system(cmd)

    return image_warped, affine_matrix, warp, inverse_warped, inverse_warp


def ants_combine_transform(in_file, transforms_list, reference):
    import os
    import os.path as op

    out_warp = op.abspath('out_warp.nii.gz')

    transforms = ""
    for trans in transforms_list:
        transforms += " " + trans
    cmd = 'antsApplyTransforms -o [out_warp.nii.gz,1] -i %s -r %s -t %s' \
          % (in_file, reference, transforms)
    os.system(cmd)

    return out_warp


def dwi_container_from_filename(bids_dwi_filename):
    """ Generate subjects/sub-<participant_id>/ses-<session_id> folder
    from BIDS filename.
    """
    import re
    from os.path import join
    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)_', bids_dwi_filename)

    if m is None:
        raise ValueError(
            'Input filename is not in a BIDS or CAPS compliant format. '
            + 'It does not contain the subject and session information.')

    subject = m.group(1)
    session = m.group(2)

    return join('subjects', subject, session)


def rename_into_caps(in_bids_dwi,
                     fname_dwi, fname_bval, fname_bvec, fname_brainmask):
    """
    Rename the outputs of the pipelines into CAPS format namely:
    <source_file>_space-T1w_preproc[.nii.gz|bval|bvec]

    Args:
        in_bids_dwi (str): Input BIDS DWI to extract the <source_file>
        fname_dwi (str): Preprocessed DWI file.
        fname_bval (str): Preprocessed bval.
        fname_bvec (str): Preprocessed bvec.
        fname_brainmask (str): B0 mask.

    Returns:
        The different outputs in CAPS format
    """
    from nipype.utils.filemanip import split_filename
    from nipype.interfaces.utility import Rename
    import os

    # Extract <source_file> in format sub-CLNC01_ses-M00[_acq-label]_dwi
    _, source_file_dwi, _ = split_filename(in_bids_dwi)

    # Extract base path from fname:
    base_dir_dwi, _, _ = split_filename(fname_dwi)
    base_dir_bval, _, _ = split_filename(fname_bval)
    base_dir_bvec, _, _ = split_filename(fname_bvec)
    base_dir_brainmask, _, _ = split_filename(fname_brainmask)

    # Rename into CAPS DWI:
    rename_dwi = Rename()
    rename_dwi.inputs.in_file = fname_dwi
    rename_dwi.inputs.format_string = source_file_dwi + "_space-T1w_preproc.nii.gz"
    out_caps_dwi = rename_dwi.run()

    # Rename into CAPS bval:
    rename_bval = Rename()
    rename_bval.inputs.in_file = fname_bval
    rename_bval.inputs.format_string = source_file_dwi + "_space-T1w_preproc.bval"
    out_caps_bval = rename_bval.run()

    # Rename into CAPS bvec:
    rename_bvec = Rename()
    rename_bvec.inputs.in_file = fname_bvec
    rename_bvec.inputs.format_string = source_file_dwi + "_space-T1w_preproc.bvec"
    out_caps_bvec = rename_bvec.run()

    # Rename into CAPS DWI:
    rename_brainmask = Rename()
    rename_brainmask.inputs.in_file = fname_brainmask
    rename_brainmask.inputs.format_string = source_file_dwi + "_space-T1w_brainmask.nii.gz"
    out_caps_brainmask = rename_brainmask.run()

    return out_caps_dwi.outputs.out_file, out_caps_bval.outputs.out_file, \
        out_caps_bvec.outputs.out_file, out_caps_brainmask.outputs.out_file


def print_begin_pipeline(in_bids_or_caps_file):
    """
    """
    from clinica.utils.stream import cprint
    import re
    import datetime
    from colorama import Fore

    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)',
                  in_bids_or_caps_file)
    if m is None:
        raise ValueError(
            'Input filename is not in a BIDS or CAPS compliant format.')
    now = datetime.datetime.now().strftime('%H:%M:%S')

    cprint('%s[%s]%s Running pipeline for %s...' % (
        Fore.BLUE, now, Fore.RESET, m.group(0)))


def print_end_pipeline(in_bids_or_caps_file, final_file):
    """
    """
    from clinica.utils.stream import cprint
    import re
    import datetime
    from colorama import Fore

    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)',
                  in_bids_or_caps_file)
    if m is None:
        raise ValueError(
            'Input filename is not in a BIDS or CAPS compliant format.')
    now = datetime.datetime.now().strftime('%H:%M:%S')

    cprint('%s[%s]%s ...%s has completed.' % (
        Fore.GREEN, now, Fore.RESET, m.group(0)))



# coding: utf8

"""
Describe files to grab, to use with inputs.clinica_file_reader() and inputs.clinica_group_reader()

"""

""" T1w """

# BIDS

T1W_NII = {'pattern': 'sub-*_ses-*_t1w.nii*',
           'description': 'T1w MRI'}

# FreeSurfer

T1_FS_WM = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/mri/wm.seg.mgz',
            'description': 'segmentation of white matter (mri/wm.seg.mgz).',
            'needed_pipeline': 't1-freesurfer'}

T1_FS_BRAIN = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/mri/brain.mgz',
               'description': ' extracted brain from T1w MRI (mri/brain.mgz).',
               'needed_pipeline': 't1-freesurfer'}

T1_FS_ORIG_NU = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/mri/orig_nu.mgz',
                 'description': 'intensity normalized volume generated after correction for'
                                ' non-uniformity in FreeSurfer (mri/orig_nu.mgz).',
                 'needed_pipeline': 't1-freesurfer'}

T1_FS_WM_SURF_R = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/surf/rh.white',
                   'description': 'right white matter/gray matter border surface (rh.white).',
                   'needed_pipeline': 't1-freesurfer'}

T1_FS_WM_SURF_L = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/surf/lh.white',
                   'description': 'left white matter/gray matter border surface (lh.white).',
                   'needed_pipeline': 't1-freesurfer'}

T1_FS_DESTRIEUX = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/mri/aparc.a2009s+aseg.mgz',
                   'description': 'Destrieux-based segmentation (mri/aparc.a2009s+aseg.mgz).',
                   'needed_pipeline': 't1-freesurfer'}

T1_FS_DESTRIEUX_PARC_L = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/label/lh.aparc.a2009s.annot',
                          'description': 'left hemisphere surface-based Destrieux parcellation (label/lh.aparc.a2009s.annot).',
                          'needed_pipeline': 't1-freesurfer'}

T1_FS_DESTRIEUX_PARC_R = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/label/rh.aparc.a2009s.annot',
                          'description': 'right hemisphere surface-based Destrieux parcellation (label/rh.aparc.a2009s.annot).',
                          'needed_pipeline': 't1-freesurfer'}

T1_FS_DESIKAN = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/mri/aparc+aseg.mgz',
                 'description': 'Desikan-based segmentation (mri/aparc.a2009s+aseg.mgz).',
                 'needed_pipeline': 't1-freesurfer'}

T1_FS_DESIKAN_PARC_L = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/label/lh.aparc.annot',
                        'description': 'left hemisphere surface-based Desikan parcellation (label/lh.aparc.annot).',
                        'needed_pipeline': 't1-freesurfer'}

T1_FS_DESIKAN_PARC_R = {'pattern': 't1/freesurfer_cross_sectional/sub-*_ses-*/label/rh.aparc.annot',
                        'description': 'right hemisphere surface-based Desikan parcellation (label/rh.aparc.annot).',
                        'needed_pipeline': 't1-freesurfer'}


# T1-Volume

def t1_volume_native_tpm(tissue_number):
    from .spm import INDEX_TISSUE_MAP
    import os
    information = {
        'pattern': os.path.join('t1', 'spm', 'segmentation', 'native_space',
                                '*_*_T1w_segm-' + INDEX_TISSUE_MAP[tissue_number] + '_probability.nii*'),
        'description': 'Tissue probability map ' + INDEX_TISSUE_MAP[tissue_number] + ' in native space',
        'needed_pipeline': 't1-volume-tissue-segmentation'
    }
    return information


def t1_volume_dartel_input_tissue(tissue_number):
    from .spm import INDEX_TISSUE_MAP
    import os
    information = {
        'pattern': os.path.join('t1', 'spm', 'segmentation', 'dartel_input',
                                '*_*_T1w_segm-' + INDEX_TISSUE_MAP[tissue_number] + '_dartelinput.nii*'),
        'description': 'Dartel input for tissue probability map ' + INDEX_TISSUE_MAP[tissue_number] + ' from T1w MRI',
        'needed_pipeline': 't1-volume-tissue-segmentation'
    }
    return information


def t1_volume_native_tpm_in_mni(tissue_number, modulation):
    from .spm import INDEX_TISSUE_MAP
    import os
    if modulation:
        pattern_modulation = 'on'
        description_modulation = "with"
    else:
        pattern_modulation = 'off'
        description_modulation = "without"
    information = {
        'pattern': os.path.join('t1', 'spm', 'segmentation', 'normalized_space',
                                '*_*_T1w_segm-' + INDEX_TISSUE_MAP[tissue_number] + '_space-Ixi549Space' +
                                '_modulated-' + pattern_modulation + '_probability.nii*'),
        'description': 'Tissue probability map ' + INDEX_TISSUE_MAP[tissue_number] + ' based on native MRI ' +
                       'in MNI space (Ixi549) ' + description_modulation + ' modulation.',
        'needed_pipeline': 't1-volume-tissue-segmentation'}
    return information


def t1_volume_template_tpm_in_mni(group_label, tissue_number, modulation):
    from .spm import INDEX_TISSUE_MAP
    import os
    if modulation:
        pattern_modulation = 'on'
        description_modulation = "with"
    else:
        pattern_modulation = 'off'
        description_modulation = "without"
    information = {
        'pattern': os.path.join('t1', 'spm', 'dartel', 'group-' + group_label,
                                '*_T1w_segm-' + INDEX_TISSUE_MAP[tissue_number] + '_space-Ixi549Space_modulated-' +
                                pattern_modulation + '_probability.nii*'),
        'description': 'Tissue probability map ' + INDEX_TISSUE_MAP[tissue_number] + ' based on ' + group_label +
                       ' template in MNI space (Ixi549) ' + description_modulation + ' modulation.',
        'needed_pipeline': 't1-volume'
    }
    return information


def t1_volume_deformation_to_template(group_label):
    import os
    information = {
        'pattern': os.path.join('t1', 'spm', 'dartel', 'group-' + group_label,
                                'sub-*_ses-*_T1w_target-' + group_label + '_transformation-forward_deformation.nii*'),
        'description': 'Deformation from native space to group template ' + group_label + ' space.',
        'needed_pipeline': 't1-volume-create-dartel'
    }
    return information


def t1_volume_i_th_iteration_group_template(group_label, i):
    import os
    information = {
        'pattern': os.path.join('group-' + group_label, 't1',
                                'group-' + group_label + '_iteration-' + str(i) + '_template.nii*'),
        'description': 'Iteration #' + str(i) + ' of Dartel template ' + group_label,
        'needed_pipeline': 't1-volume or t1-volume-create-dartel'
    }
    return information


def t1_volume_final_group_template(group_label):
    import os
    information = {
        'pattern': os.path.join('group-' + group_label, 't1',
                                'group-' + group_label + '_template.nii*'),
        'description': 'T1w template file of group ' + group_label,
        'needed_pipeline': 't1-volume or t1-volume-create-dartel'
    }
    return information


""" DWI """

# BIDS

DWI_NII = {'pattern': 'dwi/sub-*_ses-*_dwi.nii*',
           'description': 'DWI NIfTI'}

DWI_JSON = {'pattern': 'dwi/sub-*_ses-*_dwi.json',
            'description': 'DWI JSON file'}

DWI_BVAL = {'pattern': 'dwi/sub-*_ses-*_dwi.bval',
            'description': 'bval files'}

DWI_BVEC = {'pattern': 'dwi/*_dwi.bvec',
            'description': 'bvec files'}

FMAP_PHASEDIFF_JSON = {'pattern': 'fmap/sub-*_ses-*_phasediff.json',
                       'description': 'phasediff JSON file'}

FMAP_PHASEDIFF_NII = {'pattern': 'fmap/sub-*_ses-*_phasediff.nii*',
                      'description': 'phasediff NIfTI volume'}

FMAP_MAGNITUDE1_NII = {'pattern': 'fmap/sub-*_ses-*_magnitude1.nii*',
                       'description': 'magnitude1 file'}

# CAPS

DWI_PREPROC_NII = {'pattern': 'dwi/preprocessing/sub-*_ses-*_dwi_space-*_preproc.nii*',
                   'description': 'preprocessed DWI',
                   'needed_pipeline': 'dwi-preprocessing-using-t1 or dwi-preprocessing-using-fieldmap'}

DWI_PREPROC_BRAINMASK = {'pattern': 'dwi/preprocessing/sub-*_ses-*_dwi_space-*_brainmask.nii*',
                         'description': 'b0 brainmask',
                         'needed_pipeline': 'dwi-preprocessing-using-t1 or dwi-preprocessing-using-fieldmap'}

DWI_PREPROC_BVEC = {'pattern': 'dwi/preprocessing/sub-*_ses-*_dwi_space-*_preproc.bvec',
                    'description': 'preprocessed bvec',
                    'needed_pipeline': 'dwi-preprocessing-using-t1 or dwi-preprocessing-using-fieldmap'}

DWI_PREPROC_BVAL = {'pattern': 'dwi/preprocessing/*_dwi_space-*_preproc.bval',
                    'description': 'preprocessed bval',
                    'needed_pipeline': 'dwi-preprocessing-using-t1 or dwi-preprocessing-using-fieldmap'}

""" fMRI """

# BIDS
FMRI_BOLD_NII = {'pattern': 'func/sub-*_ses-*_bold.nii*',
                 'description': 'blood-oxygen-level-dependent (BOLD) image'}

FMRI_BOLD_JSON = {'pattern': 'func/sub-*_ses-*_bold.json',
                  'description': 'blood-oxygen-level-dependent (BOLD) JSON file'}

""" PET """

# BIDS

PET_FDG_NII = {'pattern': 'pet/sub-*_ses-*_acq-fdg_pet.nii*',
               'description': 'FDG-PET data'}

PET_FDG_JSON = {'pattern': 'pet/sub-*_ses-*_acq-fdg_pet.json',
                'description': 'JSON file describing the point spread function (PSF) in FDG PET.'}

PET_AV45_NII = {'pattern': 'pet/sub-*_ses-*_acq-av45_pet.nii*',
                'description': 'AV45-PET data'}

PET_AV45_JSON = {'pattern': 'pet/sub-*_ses-*_acq-av45_pet.json',
                 'description': 'JSON file describing the point spread function (PSF) in AV45 PET.'}



# coding: utf8

import clinica.engine as ce


class T1VolumeTissueSegmentationCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 't1-volume-tissue-segmentation'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = 'Tissue segmentation, bias correction and spatial normalization to MNI space' \
                            + ' of T1w images with SPM:\nhttp://clinica.run/doc/Pipelines/T1_Volume/'

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments()
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-t", "--tissue_classes",
                              metavar='', nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help="Tissue classes (1: gray matter (GM), 2: white matter (WM), "
                                   "3: cerebrospinal fluid (CSF), 4: bone, 5: soft-tissue, 6: background) to save "
                                   "(default: GM, WM and CSF i.e. --tissue_classes 1 2 3).")
        advanced.add_argument("-dt", "--dartel_tissues",
                              metavar='', nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help='Tissues to use for DARTEL template calculation '
                                   '(default: GM, WM and CSF i.e. --dartel_tissues 1 2 3).')
        advanced.add_argument("-tpm", "--tissue_probability_maps",
                              metavar='TissueProbabilityMap.nii', default=None,
                              help='Tissue probability maps to use for segmentation '
                                   '(default: TPM.nii from SPM software).')
        advanced.add_argument("-dswu", "--dont_save_warped_unmodulated",
                              action='store_true', default=False,
                              help="Do not save warped unmodulated images for tissues specified "
                                   "in --tissue_classes flag.")
        advanced.add_argument("-swm", "--save_warped_modulated",
                              action='store_true', default=False,
                              help="Save warped modulated images for tissues specified in --tissue_classes flag.")

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from .t1_volume_tissue_segmentation_pipeline import T1VolumeTissueSegmentation
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
            'tissue_classes': args.tissue_classes,
            'dartel_tissues': args.dartel_tissues,
            'tissue_probability_maps': args.tissue_probability_maps,
            'save_warped_unmodulated': not args.dont_save_warped_unmodulated,
            'save_warped_modulated': args.save_warped_modulated,
        }
        pipeline = T1VolumeTissueSegmentation(
            bids_directory=self.absolute_path(args.bids_directory),
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8


"""This module contains utilities to check dependencies of the different
neuroimaging tools."""


def is_binary_present(binary):
    """
    Check if a binary is present.

    This function checks if the program is present. Do not use this function
    with a binary GUI, it will open the GUI.

    Taken from:
    https://stackoverflow.com/questions/11210104/check-if-a-program-exists-from-a-python-script

    Args:
        binary (str): Name of the program.

    Returns:
        True if the binary is present, False otherwise.
    """
    import subprocess
    import os

    try:
        devnull = open(os.devnull)
        subprocess.Popen([binary], stdout=devnull, stderr=devnull).communicate()
    except OSError as e:
        if e.errno == os.errno.ENOENT:
            return False
    return True


def check_environment_variable(environment_variable, software_name):
    import os
    from colorama import Fore
    from .exceptions import ClinicaMissingDependencyError

    content_var = os.environ.get(environment_variable, '')
    if not content_var:
        raise ClinicaMissingDependencyError(
            '%s\n[Error] Clinica could not find %s software: the %s variable is not set.%s'
            % (Fore.RED, software_name, environment_variable, Fore.RESET))
    if not os.path.isdir(content_var):
        raise ClinicaMissingDependencyError(
            '%s\n[Error] The %s environment variable you gave is not a folder (content: %s).%s'
            % (Fore.RED, environment_variable, content_var, Fore.RESET))
    return content_var


def check_software_requirements(current_version, version_requirements, software_name):
    from distutils.version import LooseVersion
    from string import punctuation
    from colorama import Fore
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    comparison_operator = ''.join([c for c in version_requirements if c in punctuation.replace('.', '')])
    required_version = version_requirements.replace(comparison_operator, '')

    satisfy_version = eval('LooseVersion(\'%s\') %s LooseVersion(\'%s\')' %
                           (current_version, comparison_operator, required_version))

    if not satisfy_version:
        raise ClinicaMissingDependencyError(
            '%s\n[Error] Your %s version (%s) does not version requirements (%s).%s' %
            (Fore.RED, software_name, current_version, version_requirements, Fore.RESET))


def check_ants(version_requirements=None):
    """Check ANTs software."""
    from colorama import Fore
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    check_environment_variable('ANTSPATH', 'ANTs')

    list_binaries = ['N4BiasFieldCorrection', 'antsRegistrationSyNQuick.sh']
    for binary in list_binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                '%s\n[Error] Clinica could not find ANTs software: the %s command is not present '
                'in your PATH environment.%s' % (Fore.RED, binary, Fore.RESET))


def check_freesurfer(version_requirements=None):
    """Check FreeSurfer software."""
    from colorama import Fore
    import nipype.interfaces.freesurfer as freesurfer
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    check_environment_variable('FREESURFER_HOME', 'FreeSurfer')

    list_binaries = ['mri_convert', 'recon-all']
    for binary in list_binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                '%s\n[Error] Clinica could not find FreeSurfer software: the %s command is not present in your PATH '
                'environment: did you have the line source ${FREESURFER_HOME}/'
                'SetUpFreeSurfer.sh in your configuration file?%s' % (Fore.RED, binary, Fore.RESET))


def check_fsl(version_requirements=None):
    """Check FSL software."""
    import nipype.interfaces.fsl as fsl
    from colorama import Fore
    from clinica.utils.exceptions import ClinicaMissingDependencyError
    from clinica.utils.stream import cprint

    check_environment_variable('FSLDIR', 'FSL')

    try:
        if fsl.Info.version().split(".") < ['5', '0', '5']:
            raise ClinicaMissingDependencyError(
                '%sFSL version must be greater than 5.0.5%s'
                % (Fore.RED, Fore.RESET))
    except Exception as e:
        cprint(str(e))

    list_binaries = ['bet', 'flirt', 'fast', 'first']
    for binary in list_binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                '%s\n[Error] Clinica could not find FSL software: the %s command is not present in your '
                'PATH environment.%s' % (Fore.RED, binary, Fore.RESET))


def check_mrtrix(version_requirements=None):
    """Check MRtrix software."""
    from colorama import Fore
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    check_environment_variable('MRTRIX_HOME', 'MRtrix')

    list_binaries = ['transformconvert', 'mrtransform', 'dwi2response', 'tckgen']
    for binary in list_binaries:
        if not is_binary_present(binary):
            raise ClinicaMissingDependencyError(
                '%s\n[Error] Clinica could not find MRtrix software: the %s command is not present in your '
                'PATH environment.%s' % (Fore.RED, binary, Fore.RESET))


def check_spm(version_requirements=None):
    """Check SPM software."""
    check_environment_variable('SPM_HOME', 'SPM')

    # list_binaries = ['matlab']
    # for binary in list_binaries:
    #     if not is_binary_present(binary):
    #         raise RuntimeError(
    #             '%s from SPM Software is not present in your '
    #             'PATH environment.' % binary)


def check_matlab():
    """Check Matlab toolbox."""
    from colorama import Fore
    from clinica.utils.exceptions import ClinicaMissingDependencyError

    if not is_binary_present("matlab"):
        raise ClinicaMissingDependencyError(
            '%sMatlab was not found in PATH environment. Did you add it?%s'
            % (Fore.RED, Fore.RESET))


def check_cat12():
    """
    Check installation of CAT12 (mostly used to provide atlases)
    """
    import os
    from colorama import Fore
    from .exceptions import ClinicaMissingDependencyError
    from .spm import check_spm_home

    spm_home = check_spm_home()

    cat12 = os.path.join(spm_home, 'toolbox', 'cat12')
    if not os.path.exists(cat12):
        raise ClinicaMissingDependencyError(
            '%s\n[Error] CAT12 is not installed in your SPM folder. It should be in %s.%s'
            % (Fore.RED, cat12, Fore.RESET))
    return cat12


def verify_cat12_atlases(atlas_list):
    """ If the user wants to use any of the atlases of CAT12 and has not installed it, we just remove it from the list
        of the computed atlases

    :param atlas_list: list of atlases given by the user
    :return: atlas_list updated with elements removed if CAT12 is not found, or return atlas_list if
             CAT12 has been found. If none of the CAT12 atlases are in atlas_list, function returns atlas_list
    """
    import sys
    from colorama import Fore
    from time import sleep
    from .exceptions import ClinicaMissingDependencyError

    cat12_atlases = ['Hammers', 'Neuromorphometrics', 'LPBA40']
    if any(atlas in cat12_atlases for atlas in atlas_list):
        try:
            check_cat12()
            atlas_list_updated = atlas_list
        except ClinicaMissingDependencyError:
            print(Fore.YELLOW + '[Warning] CAT12 is not installed in your system. Atlas statistics computed at the '
                  + 'end of this pipeline will not include any of Hammers, Neuromorphometrics and LPBA40'
                  + ' atlases.' + Fore.RESET)
            print('Pipeline will start in 15 sec. You can cancel it right now using Ctrl + C')
            print('You can download CAT12 at this address: ' + Fore.BLUE + 'http://www.neuro.uni-jena.de/cat/'
                  + Fore.RESET)
            atlas_list_updated = [atl for atl in atlas_list if atl not in cat12_atlases]
            try:
                sleep(15)
            except KeyboardInterrupt:
                print('\nClinica is now exiting.')
                sys.exit()
    else:
        atlas_list_updated = atlas_list
    return atlas_list_updated



# coding: utf8

import clinica.pipelines.engine as cpe


class StatisticsSurface(cpe.Pipeline):
    """
    Based on the Matlab toolbox [SurfStat](http://www.math.mcgill.ca/keith/surfstat/), which performs statistical
    analyses of univariate and multivariate surface and volumetric data using the generalized linear model (GLM),
    this pipeline performs analyses including group comparison and correlation with surface-based features e.g.
    cortical thickness from t1-freesurfer or map of activity from PET data from pet-surface pipeline.

    TODO: Refactor StatisticsSurface
        [ ] build_input_node
            [ ] Remove current read_parameters_node
            [ ] With the help of statistics_surface_utils.py::check_inputs, use new clinica_file_reader function
                to check and extract surface-based features
            [ ] Delete statistics_surface_utils.py::check_inputs function
            [ ] Move statistics_surface_cli.py checks of input data in this method
            [ ] Handle overwrite case
            [ ] Display participants, covariates and info regarding the GLM.
        [ ] build_core_nodes
            [ ] Use surfstat.inputs.full_width_at_half_maximum = self.parameter['full_width_at_half_maximum']
                instead of connecting read_parameters_node.inputs.full_width_at_half_maximum
                to surfstat.inputs.full_width_at_half_maximum
            [ ] Repeat for other keys
            [ ] Use working directory
        [ ] build_output_node
            [ ] Remove path_to_matscript and freesurfer_home: it should be set in runmatlab function
            [ ] Copy results from <WD> to <CAPS>
        [ ] Clean/adapt statistics_surface_utils.py

    Args:
        caps_directory: str, the output folder of recon-all which will contain the result files: ?h.thickness.fwhm**.mgh.
        tsv_file: str, Path to the tsv containing the information for GLM.
        design_matrix: str, the linear model that fits into the GLM, for example '1+group'.
        contrast: string, the contrast matrix for GLM, if the factor you choose is categorized variable, clinica_surfstat will create two contrasts,
                  for example, contrast = 'Label', this will create contrastpos = Label.AD - Label.CN, contrastneg = Label.CN - Label.AD; if the fac-
                  tory that you choose is a continuous factor, clinica_surfstat will just create one contrast, for example, contrast = 'Age', but note,
                  the string name that you choose should be exactly the same with the columns names in your subjects_visits_tsv.
        str_format: string, the str_format which uses to read your tsv file, the typy of the string should corresponds exactly with the columns in the tsv file.
            Defaut parameters, we set these parameters to be some default values, but you can also set it by yourself:
        group_label: current group name for this analysis
        glm_type: based on the hypothesis, you should define one of the glm types, "group_comparison", "correlation"
        full_width_at_half_maximum: fwhm for the surface smoothing, default is 20, integer.
        threshold_uncorrected_pvalue: threshold to display the uncorrected Pvalue, float, default is 0.001.
        threshold_corrected_pvalue: the threshold to display the corrected cluster, default is 0.05, float.
        cluster_threshold: threshold to define a cluster in the process of cluster-wise correction, default is 0.001, float.
        working_directory: define where to put the infomartion of the nipype workflow.
        n_procs: define how many cores to run this workflow.

    Returns:
        A clinica pipeline object containing the StatisticsSurface pipeline.
    """
    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from .statistics_surface_utils import get_t1_freesurfer_custom_file
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.group import check_group_label

        if 'custom_file' not in self.parameters.keys():
            self.parameters['custom_file'] = get_t1_freesurfer_custom_file()
        if 'feature_label' not in self.parameters.keys():
            self.parameters['feature_label'] = 'ct',
        if 'full_width_at_half_maximum' not in self.parameters.keys():
            self.parameters['full_width_at_half_maximum'] = 20
        if 'threshold_uncorrected_pvalue' not in self.parameters.keys():
            self.parameters['threshold_uncorrected_pvalue'] = 0.001
        if 'threshold_corrected_pvalue' not in self.parameters.keys():
            self.parameters['threshold_corrected_pvalue'] = 0.05,
        if 'cluster_threshold' not in self.parameters.keys():
            self.parameters['cluster_threshold'] = 0.001,

        check_group_label(self.parameters['group_label'])
        if self.parameters['glm_type'] not in ['group_comparison', 'correlation']:
            raise ClinicaException("The glm_type you specified is wrong: it should be group_comparison or "
                                   "correlation (given value: %s)." % self.parameters['glm_type'])
        if self.parameters['full_width_at_half_maximum'] not in [0, 5, 10, 15, 20]:
            raise ClinicaException(
                "FWHM for the surface smoothing you specified is wrong: it should be 0, 5, 10, 15 or 20 "
                "(given value: %s)." % self.parameters['full_width_at_half_maximum'])
        if self.parameters['threshold_uncorrected_pvalue'] < 0 or self.parameters['threshold_uncorrected_pvalue'] > 1:
            raise ClinicaException("Uncorrected p-value threshold should be a lower than 1 "
                                   "(given value: %s)." % self.parameters['threshold_uncorrected_pvalue'])
        if self.parameters['threshold_corrected_pvalue'] < 0 or self.parameters['threshold_corrected_pvalue'] > 1:
            raise ClinicaException("Corrected p-value threshold should be between 0 and 1 "
                                   "(given value: %s)." % self.parameters['threshold_corrected_pvalue'])
        if self.parameters['cluster_threshold'] < 0 or self.parameters['cluster_threshold'] > 1:
            raise ClinicaException("Cluster threshold should be between 0 and 1 "
                                   "(given value: %s)." % self.parameters['cluster_threshold'])

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """

        return ['design_matrix',
                'contrast',
                'str_format',
                'group_label',
                'glm_type',
                'surface_file',
                'full_width_at_half_maximum',
                'threshold_uncorrected_pvalue',
                'threshold_corrected_pvalue',
                'cluster_threshold',
                'feature_label']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """

        return []

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        read_parameters_node = npe.Node(name="LoadingCLIArguments",
                                        interface=nutil.IdentityInterface(
                                            fields=self.get_input_fields(),
                                            mandatory_inputs=True))
        read_parameters_node.inputs.design_matrix = self.parameters['design_matrix']
        read_parameters_node.inputs.contrast = self.parameters['contrast']
        read_parameters_node.inputs.str_format = self.parameters['str_format']
        read_parameters_node.inputs.group_label = self.parameters['group_label']
        read_parameters_node.inputs.glm_type = self.parameters['glm_type']
        read_parameters_node.inputs.surface_file = self.parameters['custom_file']
        read_parameters_node.inputs.full_width_at_half_maximum = self.parameters['full_width_at_half_maximum']
        read_parameters_node.inputs.threshold_uncorrected_pvalue = self.parameters['threshold_uncorrected_pvalue']
        read_parameters_node.inputs.threshold_corrected_pvalue = self.parameters['threshold_corrected_pvalue']
        read_parameters_node.inputs.cluster_threshold = self.parameters['cluster_threshold']
        read_parameters_node.inputs.feature_label = self.parameters['feature_label']

        self.connect([
            (read_parameters_node, self.input_node, [('design_matrix',                'design_matrix')]),  # noqa
            (read_parameters_node, self.input_node, [('surface_file',                 'surface_file')]),  # noqa
            (read_parameters_node, self.input_node, [('contrast',                     'contrast')]),  # noqa
            (read_parameters_node, self.input_node, [('str_format',                   'str_format')]),  # noqa
            (read_parameters_node, self.input_node, [('group_label',                  'group_label')]),  # noqa
            (read_parameters_node, self.input_node, [('glm_type',                     'glm_type')]),  # noqa
            (read_parameters_node, self.input_node, [('full_width_at_half_maximum',   'full_width_at_half_maximum')]),  # noqa
            (read_parameters_node, self.input_node, [('threshold_uncorrected_pvalue', 'threshold_uncorrected_pvalue')]),  # noqa
            (read_parameters_node, self.input_node, [('threshold_corrected_pvalue',   'threshold_corrected_pvalue')]),  # noqa
            (read_parameters_node, self.input_node, [('cluster_threshold',            'cluster_threshold')]),  # noqa
            (read_parameters_node, self.input_node, [('feature_label',                'feature_label')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """

        pass

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """

        import clinica.pipelines.statistics_surface.statistics_surface_utils as utils
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces.io import JSONFileSink

        # Node to fetch the input variables
        data_prep = npe.Node(name='inputnode',
                             interface=nutil.Function(
                                 input_names=['input_directory', 'subjects_visits_tsv', 'group_label', 'glm_type'],
                                 output_names=['path_to_matscript', 'surfstat_input_dir', 'output_directory',
                                               'freesurfer_home', 'out_json'],
                                 function=utils.prepare_data))
        data_prep.inputs.input_directory = self.caps_directory
        data_prep.inputs.subjects_visits_tsv = self.tsv_file

        # Node to wrap the SurfStat matlab script
        surfstat = npe.Node(name='surfstat',
                            interface=nutil.Function(
                                input_names=['input_directory',
                                             'output_directory',
                                             'subjects_visits_tsv',
                                             'design_matrix',
                                             'contrast',
                                             'str_format',
                                             'glm_type',
                                             'group_label',
                                             'freesurfer_home',
                                             'surface_file',
                                             'path_to_matscript',
                                             'full_width_at_half_maximum',
                                             'threshold_uncorrected_pvalue',
                                             'threshold_corrected_pvalue',
                                             'cluster_threshold',
                                             'feature_label'],
                                output_names=['out_images'],
                                function=utils.runmatlab))
        surfstat.inputs.subjects_visits_tsv = self.tsv_file

        # Node to create the dictionary for JSONFileSink
        jsondict = npe.Node(name='Jsondict',
                            interface=nutil.Function(
                                input_names=['glm_type', 'design_matrix', 'str_format', 'contrast', 'group_label', 'full_width_at_half_maximum',
                                             'threshold_uncorrected_pvalue', 'threshold_corrected_pvalue', 'cluster_threshold'],
                                output_names=['json_dict'],
                                function=utils.json_dict_create))

        # Node to write the GLM information into a JSON file
        jsonsink = npe.Node(JSONFileSink(input_names=['out_file']), name='jsonsinker')

        # Connection
        # ==========
        self.connect([
            (self.input_node, data_prep, [('group_label', 'group_label')]),  # noqa
            (self.input_node, data_prep, [('glm_type', 'glm_type')]),  # noqa
            (self.input_node, surfstat, [('design_matrix', 'design_matrix')]),  # noqa
            (self.input_node, surfstat, [('contrast', 'contrast')]),  # noqa
            (self.input_node, surfstat, [('str_format', 'str_format')]),  # noqa
            (self.input_node, surfstat, [('glm_type', 'glm_type')]),  # noqa
            (self.input_node, surfstat, [('group_label', 'group_label')]),  # noqa
            (self.input_node, surfstat, [('full_width_at_half_maximum', 'full_width_at_half_maximum')]),  # noqa
            (self.input_node, surfstat, [('threshold_uncorrected_pvalue', 'threshold_uncorrected_pvalue')]),  # noqa
            (self.input_node, surfstat, [('threshold_corrected_pvalue', 'threshold_corrected_pvalue')]),  # noqa
            (self.input_node, surfstat, [('cluster_threshold', 'cluster_threshold')]),  # noqa
            (self.input_node, surfstat, [('surface_file', 'surface_file')]),  # noqa
            (self.input_node, surfstat, [('feature_label', 'feature_label')]),
            (data_prep, surfstat, [('surfstat_input_dir', 'input_directory')]),  # noqa
            (data_prep, surfstat, [('path_to_matscript', 'path_to_matscript')]),  # noqa
            (data_prep, surfstat, [('output_directory', 'output_directory')]),  # noqa
            (data_prep, surfstat, [('freesurfer_home', 'freesurfer_home')]),  # noqa
            (self.input_node, jsondict, [('glm_type', 'glm_type')]),  # noqa
            (self.input_node, jsondict, [('design_matrix', 'design_matrix')]),  # noqa
            (self.input_node, jsondict, [('str_format', 'str_format')]),  # noqa
            (self.input_node, jsondict, [('contrast', 'contrast')]),  # noqa
            (self.input_node, jsondict, [('group_label', 'group_label')]),  # noqa
            (self.input_node, jsondict, [('full_width_at_half_maximum', 'full_width_at_half_maximum')]),  # noqa
            (self.input_node, jsondict, [('threshold_uncorrected_pvalue', 'threshold_uncorrected_pvalue')]),  # noqa
            (self.input_node, jsondict, [('threshold_corrected_pvalue', 'threshold_corrected_pvalue')]),  # noqa
            (self.input_node, jsondict, [('cluster_threshold', 'cluster_threshold')]),  # noqa
            (data_prep, jsonsink, [('out_json', 'out_file')]),  # noqa
            (jsondict, jsonsink, [('json_dict', 'in_dict')]),  # noqa
        ])



# coding: utf8
from clinica.utils.filemanip import read_participant_tsv


def get_unique_subjects(in_subject_list, in_session_list):
    """Get unique participant IDs

    The function to read the .tsv file returns the following
    participant_id and session_id lists:
    participant1, participant1, ..., participant2, participant2, ...
    session1    , session2    , ..., session1    , session2    , ...
    This function returns a list where all participants are only selected
    once:
    participant1, participant2, ..., participant_n
    and for each participant, the list of corresponding session id
    eg.:
    participant1 -> [session1, session2]
    participant2 -> [session1]
    ...
    participant_n -> [session1, session2, session3]

    Args:
        in_subject_list (list of strings): list of participant_id
        in_session_list (list of strings): list of session_id

    Returns:
        out_unique_subject_list (list of strings): list of
            participant_id, where each participant appears only once
        out_persubject_session_list2 (list of list): list of list
            (list2) of session_id associated to any single participant
    """

    import numpy as np

    subject_array = np.array(in_subject_list)
    session_array = np.array(in_session_list)

    # The second returned element indicates for each participant_id the
    # element they correspond to in the 'unique' list. We will use this
    # to link each session_id in the repeated list of session_id to
    # their corresponding unique participant_id

    unique_subject_array, out_inverse_positions = np.unique(
        subject_array, return_inverse=True)
    out_unique_subject_list = unique_subject_array.tolist()

    subject_number = len(out_unique_subject_list)
    out_persubject_session_list2 = [
        session_array[
            out_inverse_positions == subject_index
            ].tolist() for subject_index in range(subject_number)]

    assert len(out_unique_subject_list) == len(out_persubject_session_list2), 'Problem while getting unique subjects and sessions lists'

    return out_unique_subject_list, out_persubject_session_list2


def get_subject_session_list(input_dir, ss_file=None, is_bids_dir=True, use_session_tsv=False, tsv_dir=None):
    """Parse a BIDS or CAPS directory to get the subjects and sessions.

    This function lists all the subjects and sessions based on the content of
    the BIDS or CAPS directory or (if specified) on the provided
    subject-sessions TSV file.

    Args:
        input_dir: A BIDS or CAPS directory path.
        ss_file: A subjects-sessions file (.tsv format).
        is_bids_dir: Indicates if input_dir is a BIDS or CAPS directory
        use_session_tsv (boolean): Specify if the list uses the sessions listed in the sessions.tsv files
        tsv_dir (str): if TSV file does not exist, it will be created in output_dir. If
            not specified, output_dir will be in <tmp> folder

    Returns:
        subjects: A subjects list.
        sessions: A sessions list.

    Notes:
        This is a generic method based on folder names. If your <BIDS> dataset contains e.g.:
        - sub-CLNC01/ses-M00/anat/sub-CLNC01_ses-M00_T1w.nii
        - sub-CLNC02/ses-M00/dwi/sub-CLNC02_ses-M00_dwi.{bval|bvec|json|nii}
        - sub-CLNC02/ses-M00/anat/sub-CLNC02_ses-M00_T1w.nii
        get_subject_session_list(<BIDS>, None, True) will return
        ['ses-M00', 'ses-M00'], ['sub-CLNC01', 'sub-CLNC02'].

        However, if your pipeline needs both T1w and DWI files, you will need to check
        with e.g. clinica_file_reader_function.
    """
    import os
    import tempfile
    from time import time, strftime, localtime
    import clinica.iotools.utils.data_handling as cdh

    if not ss_file:
        if tsv_dir:
            output_dir = tsv_dir
        else:
            output_dir = tempfile.mkdtemp()
        timestamp = strftime('%Y%m%d_%H%M%S', localtime(time()))
        tsv_file = 'subjects_sessions_list_%s.tsv' % timestamp
        ss_file = os.path.join(output_dir, tsv_file)

        cdh.create_subs_sess_list(
            input_dir=input_dir,
            output_dir=output_dir,
            file_name=tsv_file,
            is_bids_dir=is_bids_dir,
            use_session_tsv=use_session_tsv)

    participant_ids, session_ids = read_participant_tsv(ss_file)
    return session_ids, participant_ids



# coding: utf8

import clinica.engine as ce


class T1VolumeRegisterDartelCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 't1-volume-register-dartel'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('Inter-subject registration using Dartel (using an existing Dartel template):\n'
                             'http://clinica.run/doc/Pipelines/T1_Volume/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_id",
                                  help='User-defined identifier for the provided group of subjects.')
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-s", "--smooth",
                              nargs='+', type=int, default=[8],
                              help="A list of integers specifying the different isomorphic FWHM in millimeters "
                                   "to smooth the image (default: --smooth 8).")
        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments()
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-t", "--tissues",
                              metavar='', nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help='Tissues to create flow fields to DARTEL template '
                                   '(default: GM, WM and CSF i.e. --tissues 1 2 3).')

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from .t1_volume_register_dartel_pipeline import T1VolumeRegisterDartel
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
            'group_id': args.group_id,
            'tissues': args.tissues,
        }
        pipeline = T1VolumeRegisterDartel(
            bids_directory=self.absolute_path(args.bids_directory),
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os.path as op

from nipype.utils.filemanip import split_filename
from nipype.interfaces.base import (CommandLineInputSpec, CommandLine, traits, TraitedSpec,
                                    File, InputMultiPath, isdefined)


class DWI2TensorInputSpec(CommandLineInputSpec):
    in_file = InputMultiPath(
        File(exists=True),
        argstr='%s',
        mandatory=True,
        position=-2,
        desc='Diffusion-weighted images')
    out_filename = File(
        name_template="%s_tensor.mif",
        name_source="in_file",
        output_name="tensor",
        argstr='%s',
        desc='Output tensor filename',
        position=-1)
    encoding_file = File(
        argstr='-grad %s',
        position=2,
        desc=('Encoding file supplied as a 4xN text file with '
              'each line is in the format [ X Y Z b ], where '
              '[ X Y Z ] describe the direction of the applied '
              'gradient, and b gives the b-value in units '
              '(1000 s/mm^2). See FSL2MRTrix()'))
    ignore_slice_by_volume = traits.List(
        traits.Int,
        argstr='-ignoreslices %s',
        sep=' ',
        position=2,
        minlen=2,
        maxlen=2,
        desc=('Requires two values (i.e. [34 '
              '1] for [Slice Volume] Ignores '
              'the image slices specified '
              'when computing the tensor. '
              'Slice here means the z '
              'coordinate of the slice to be '
              'ignored.'))
    ignore_volumes = traits.List(
        traits.Int,
        argstr='-ignorevolumes %s',
        sep=' ',
        position=2,
        minlen=1,
        desc=('Requires two values (i.e. [2 5 6] for '
              '[Volumes] Ignores the image volumes '
              'specified when computing the tensor.'))
    quiet = traits.Bool(
        argstr='-quiet',
        position=1,
        desc=("Do not display information messages or progress "
              "status."))
    debug = traits.Bool(
        argstr='-debug', position=1, desc="Display debugging messages.")
    in_mask = File(
        exists=True,
        argstr='-mask %s',
        desc=('only perform computation within the specified binary'
              ' brain mask image'))


class DWI2TensorOutputSpec(TraitedSpec):
    tensor = File(
        exists=True, desc='path/name of output diffusion tensor image')


class DWI2Tensor(CommandLine):
    """
    Converts diffusion-weighted images to tensor images.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> dwi2tensor = mrt.DWI2Tensor()
    >>> dwi2tensor.inputs.in_file = 'dwi.mif'
    >>> dwi2tensor.inputs.encoding_file = 'encoding.txt'
    >>> dwi2tensor.cmdline
    'dwi2tensor -grad encoding.txt dwi.mif dwi_tensor.mif'
    >>> dwi2tensor.run()                                   # doctest: +SKIP
    """

    _cmd = 'dwi2tensor'
    input_spec = DWI2TensorInputSpec
    output_spec = DWI2TensorOutputSpec


class MRTransformInputSpec(CommandLineInputSpec):
    in_files = InputMultiPath(
        File(exists=True),
        argstr='%s',
        mandatory=True,
        position=-2,
        desc='Input images to be transformed')
    out_filename = File(
        genfile=True, argstr='%s', position=-1, desc='Output image')
    invert = traits.Bool(
        argstr='-inverse',
        position=1,
        desc="Invert the specified transform before using it")
    replace_transform = traits.Bool(
        argstr='-replace',
        position=1,
        desc="replace the current transform by that specified, rather than applying it to the current transform"
    )
    transformation_file = File(
        exists=True,
        argstr='-transform %s',
        position=1,
        desc='The transform to apply, in the form of a 4x4 ascii file.')
    linear_transform = File(
        exists=True,
        argstr='-linear %s',
        position=1,
        desc=('specify a linear transform to apply, in the form of a 3x4 or '
              '4x4 ascii file. Note the standard reverse convention is used, '
              'where the transform maps points in the template image to the '
              'moving image. Note that the reverse convention is still assumed '
              'even if no -template image is supplied'))
    template_image = File(
        exists=True,
        argstr='-template %s',
        position=1,
        desc='Reslice the input image to match the specified template image.')
    reference_image = File(
        exists=True,
        argstr='-reference %s',
        position=1,
        desc='in case the transform supplied maps from the input image onto a reference image, use this option to specify the reference. Note that this implicitly sets the -replace option.'
    )
    flip_x = traits.Bool(
        argstr='-flipx',
        position=1,
        desc="assume the transform is supplied assuming a coordinate system with the x-axis reversed relative to the MRtrix convention (i.e. x increases from right to left). This is required to handle transform matrices produced by FSL's FLIRT command. This is only used in conjunction with the -reference option."
    )
    quiet = traits.Bool(
        argstr='-quiet',
        position=1,
        desc="Do not display information messages or progress status.")
    debug = traits.Bool(
        argstr='-debug', position=1, desc="Display debugging messages.")


class MRTransformOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='the output image of the transformation')


class MRTransform(CommandLine):
    """
    Apply spatial transformations or reslice images

    Example
    -------

    >>> MRxform = MRTransform()
    >>> MRxform.inputs.in_files = 'anat_coreg.mif'
    >>> MRxform.run()                                   # doctest: +SKIP
    """

    _cmd = 'mrtransform'
    input_spec = MRTransformInputSpec
    output_spec = MRTransformOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = self.inputs.out_filename
        if not isdefined(outputs['out_file']):
            outputs['out_file'] = op.abspath(self._gen_outfilename())
        else:
            outputs['out_file'] = op.abspath(outputs['out_file'])
        return outputs

    def _gen_filename(self, name):
        if name == 'out_filename':
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_files[0])
        return name + '_MRTransform.mif'
